import numpy as np
from oasislmf.pytools.common.data import oasis_float
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO


def fm_tobin(stack, file_in, file_out, file_type, stream_type, max_sample_index):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)

    stream_agg_type = 1
    stream_info = (stream_type << 24 | stream_agg_type)
    # Write stream info byte
    np.array([stream_info], dtype="i4").tofile(file_out)
    # Write sample len byte
    np.array([max_sample_index], dtype="i4").tofile(file_out)

    curr_event_id = -1
    curr_item_id = -1
    sidx_losses = []
    sidx_loss_dtype = np.dtype([
        ("sidx", np.int32),
        ("loss", oasis_float)]
    )
    for row in data:
        event_id = row["event_id"]
        item_id = row["output_id"]
        sidx = row["sidx"]
        loss = row["loss"]
        if (event_id != curr_event_id) or (item_id != curr_item_id):
            if curr_event_id != -1:
                sidx_losses.append((0, 0))
                np.array([curr_event_id, curr_item_id], dtype=np.int32).tofile(file_out)
                np.array(sidx_losses, dtype=sidx_loss_dtype).tofile(file_out)
            curr_event_id = event_id
            curr_item_id = item_id
            sidx_losses = []
        if sidx <= max_sample_index:
            sidx_losses.append((sidx, loss))
    if curr_event_id != -1:
        sidx_losses.append((0, 0))
        np.array([curr_event_id, curr_item_id], dtype=np.int32).tofile(file_out)
        np.array(sidx_losses, dtype=sidx_loss_dtype).tofile(file_out)



from oasislmf.pytools.common.data import fm_policytc_headers, fm_policytc_dtype, fm_policytc_fmt


headers = fm_policytc_headers
dtype = fm_policytc_dtype
fmt = fm_policytc_fmt
cli_support = ['bintocsv', 'csvtobin']




import logging
import numba as nb
import numpy as np
from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, oasis_int, oasis_int_size, oasis_float, oasis_float_size, write_ndarray_to_fmt_csv
from oasislmf.pytools.common.event_stream import (
    GUL_STREAM_ID, LOSS_STREAM_ID, EventReader, init_streams_in, mv_read
)
from oasislmf.pytools.converters.data import TOOL_INFO

logger = logging.getLogger(__name__)


class GulReader(EventReader):
    def __init__(self, len_sample, data_dtype):
        self.logger = logger
        self.data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=data_dtype)
        self.idx = np.zeros(1, dtype=np.int64)

        read_buffer_state_dtype = np.dtype([
            ('len_sample', oasis_int),
            ('reading_losses', np.bool_),
        ])

        self.state = np.zeros(1, dtype=read_buffer_state_dtype)[0]
        self.state["len_sample"] = len_sample

    def read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id, **kwargs):
        cursor, event_id, item_id, ret = read_buffer(
            byte_mv,
            cursor,
            valid_buff,
            event_id,
            item_id,
            self.data,
            self.idx,
            self.state
        )
        return cursor, event_id, item_id, ret


@nb.njit(cache=True)
def read_buffer(byte_mv, cursor, valid_buff, event_id, item_id, data, idxs, state):
    last_event_id = event_id
    idx = idxs[0]

    def _reset_state():
        state["reading_losses"] = False

    def _update_idxs():
        idxs[0] = idx

    while cursor < valid_buff:
        if not state["reading_losses"]:
            # Read summary header
            if valid_buff - cursor >= 2 * oasis_int_size:
                event_id_new, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
                if last_event_id != 0 and event_id_new != last_event_id:
                    # New event, return to process the previous event
                    _update_idxs()
                    return cursor - oasis_int_size, last_event_id, item_id, 1
                event_id = event_id_new
                item_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
                state["reading_losses"] = True
            else:
                break  # Not enough for whole summary header

        if state["reading_losses"]:
            if valid_buff - cursor < oasis_int_size + oasis_float_size:
                break  # Not enough for whole record

            # Read sidx
            sidx, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            if sidx == 0:  # sidx == 0, end of record
                cursor += oasis_float_size  # Read extra 0 for end of record
                _reset_state()
                continue

            # Read loss
            loss, cursor = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
            data[idx]["event_id"] = event_id
            data[idx]["item_id"] = item_id
            data[idx]["sidx"] = sidx
            data[idx]["loss"] = loss
            idx += 1
            if idx >= data.shape[0]:
                # Output array is full
                _update_idxs()
                return cursor, event_id, item_id, 1
        else:
            pass  # Should never reach here

    # Update the indices
    _update_idxs()
    return cursor, event_id, item_id, 0


def gul_tocsv(stack, file_in, file_out, file_type, noheader):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    if str(file_in) == "-":
        file_in = None  # init_streams checks for None to read from sys.stdin.buffer

    streams_in, (stream_source_type, stream_agg_type, len_sample) = init_streams_in(file_in, stack)
    if stream_source_type not in [GUL_STREAM_ID, LOSS_STREAM_ID]:
        raise Exception(f"unsupported stream type {stream_source_type}, {stream_agg_type}")

    if not noheader:
        file_out.write(",".join(headers) + "\n")

    gul_reader = GulReader(len_sample=len_sample, data_dtype=dtype)

    for event_id in gul_reader.read_streams(streams_in):
        idx = gul_reader.idx
        data = gul_reader.data[:idx[0]]
        write_ndarray_to_fmt_csv(
            file_out,
            data,
            headers,
            fmt,
        )
        idx[0] = 0



from oasislmf.pytools.plt.data import SPLT_headers, SPLT_dtype, SPLT_fmt


headers = SPLT_headers
dtype = SPLT_dtype
fmt = SPLT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



"""
This file defines the data types that are loaded from the data files.
"""
import numba as nb
import numpy as np

from oasislmf.pytools.common.data import areaperil_int, oasis_float

# Footprint file formats in order of priority
fp_format_priorities = [
    'parquet_chunk', 'parquet', 'binZ', 'bin', 'csv', 'parquet_dynamic'
]

# filenames
footprint_filename = 'footprint.bin'
footprint_index_filename = 'footprint.idx'
zfootprint_filename = 'footprint.bin.z'
zfootprint_index_filename = 'footprint.idx.z'
csvfootprint_filename = 'footprint.csv'
parquetfootprint_filename = 'footprint.parquet'
parquetfootprint_chunked_dir = 'footprint_chunk'
parquetfootprint_chunked_lookup = 'footprint_lookup.parquet'
footprint_bin_lookup = 'footprint_lookup.bin'
footprint_csv_lookup = 'footprint_lookup.csv'
parquetfootprint_meta_filename = 'footprint_parquet_meta.json'
event_defintion_filename = 'event_definition.parquet'
hazard_case_filename = 'hazard_case.parquet'


FootprintHeader = nb.from_dtype(np.dtype([('num_intensity_bins', np.int32),
                                          ('has_intensity_uncertainty',
                                           np.int32)
                                          ]))

Event_dtype = np.dtype([
    ('areaperil_id', areaperil_int),
    ('intensity_bin_id', np.int32),
    ('probability', oasis_float)
])
Event = nb.from_dtype(Event_dtype)

EventDynamic = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('intensity_bin_id', np.int32),
                                       ('intensity', np.int32),
                                       ('probability', oasis_float),
                                       ('return_period', np.int32)
                                       ]))

EventIndexBin_dtype = np.dtype([
    ('event_id', np.int32),
    ('offset', np.int64),
    ('size', np.int64)
])
EventIndexBin = nb.from_dtype(EventIndexBin_dtype)

EventIndexBinZ_dtype = np.dtype([
    ('event_id', np.int32),
    ('offset', np.int64),
    ('size', np.int64),
    ('d_size', np.int64)
])
EventIndexBinZ = nb.from_dtype(EventIndexBinZ_dtype)

Index_type = nb.from_dtype(np.dtype([('start', np.int64),
                                     ('end', np.int64)
                                     ]))

Event_defintion = nb.from_dtype(np.dtype([('section_id', np.int32),
                                          ('return_period', np.int32),
                                          ('rp_from', np.int32),
                                          ('rp_to', np.int32),
                                          ('interpolation', np.int32)
                                          ]))

Hazard_case = nb.from_dtype(np.dtype([('section_id', np.int32),
                                      ('areaperil_id', areaperil_int),
                                      ('return_period', np.int32),
                                      ('intensity', np.int32)
                                      ]))

Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}



from oasislmf.pytools.aal.data import ALCT_headers, ALCT_dtype, ALCT_fmt


headers = ALCT_headers
dtype = ALCT_dtype
fmt = ALCT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



from oasislmf.pytools.common.data import correlations_headers, correlations_dtype, correlations_fmt


headers = correlations_headers
dtype = correlations_dtype
fmt = correlations_fmt
cli_support = ['bintocsv', 'csvtobin']



from pathlib import Path
import numba as nb
import numpy as np
from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, generate_output_metadata, write_ndarray_to_fmt_csv
from oasislmf.pytools.common.input_files import occ_get_date, read_occurrence_bin
from oasislmf.pytools.converters.data import TOOL_INFO


def occurrence_tocsv(stack, file_in, file_out, file_type, noheader):
    @nb.njit(cache=True, error_model="numpy")
    def _get_occ_data_with_dates(occ_arr, occ_csv_dtype):
        buffer_size = DEFAULT_BUFFER_SIZE
        buffer = np.zeros(buffer_size, dtype=occ_csv_dtype)

        idx = 0
        for row in occ_arr:
            if idx >= buffer_size:
                yield buffer[:idx]
                buffer = np.zeros(buffer_size, dtype=occ_csv_dtype)
                idx = 0
            year, month, day, hour, minute = occ_get_date(row["occ_date_id"], granular_date)
            buffer[idx]["event_id"] = row["event_id"]
            buffer[idx]["period_no"] = row["period_no"]
            buffer[idx]["occ_year"] = year
            buffer[idx]["occ_month"] = month
            buffer[idx]["occ_day"] = day
            idx += 1
        yield buffer[:idx]

    @nb.njit(cache=True, error_model="numpy")
    def _get_occ_data_with_dates_gran(occ_arr, occ_csv_dtype):
        buffer_size = DEFAULT_BUFFER_SIZE
        buffer = np.zeros(buffer_size, dtype=occ_csv_dtype)

        idx = 0
        for row in occ_arr:
            if idx >= buffer_size:
                yield buffer[:idx]
                buffer = np.zeros(buffer_size, dtype=occ_csv_dtype)
                idx = 0
            year, month, day, hour, minute = occ_get_date(row["occ_date_id"], granular_date)
            buffer[idx]["event_id"] = row["event_id"]
            buffer[idx]["period_no"] = row["period_no"]
            buffer[idx]["occ_year"] = year
            buffer[idx]["occ_month"] = month
            buffer[idx]["occ_day"] = day
            buffer[idx]["occ_hour"] = hour
            buffer[idx]["occ_minute"] = minute
            idx += 1
        yield buffer[:idx]

    run_dir = Path(file_in).parent
    filename = Path(file_in).name
    if str(file_in) == "-":
        occ_arr, date_algorithm, granular_date, no_of_periods = read_occurrence_bin(
            use_stdin=True
        )
    else:
        occ_arr, date_algorithm, granular_date, no_of_periods = read_occurrence_bin(
            run_dir=run_dir,
            filename=filename
        )
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    if date_algorithm:
        occ_csv_output = [
            ("event_id", 'i4', "%d"),
            ("period_no", 'i4', "%d"),
            ("occ_year", 'i4', "%d"),
            ("occ_month", 'i4', "%d"),
            ("occ_day", 'i4', "%d"),
        ]
        if granular_date:
            occ_csv_output += [
                ("occ_hour", 'i4', "%d"),
                ("occ_minute", 'i4', "%d"),
            ]
        headers, dtype, fmt = generate_output_metadata(occ_csv_output)
        gen = _get_occ_data_with_dates(occ_arr, dtype)
        if granular_date:
            gen = _get_occ_data_with_dates_gran(occ_arr, dtype)

        if not noheader:
            file_out.write(",".join(headers) + "\n")
        for data in gen:
            write_ndarray_to_fmt_csv(file_out, data, headers, fmt)
    else:
        if not noheader:
            file_out.write(",".join(headers) + "\n")
        write_ndarray_to_fmt_csv(file_out, occ_arr, headers, fmt)



from importlib import resources
import json
from jsonschema import validate, ValidationError
import os
from pathlib import Path

from oasislmf.computation.base import ComputationStep
from oasislmf.utils.documentation.jsontomd import DefaultJsonToMarkdownGenerator, RDLS_0_2_0_JsonToMarkdownGenerator


class GenerateModelDocumentation(ComputationStep):
    """
    Generates Model Documentation from schema provided in the model config file
    """
    # Command line options
    step_params = [
        {'name': 'doc_out_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False,
         'help': 'Path to the directory in which to generate the Documentation files', 'default': '.'},
        {'name': 'doc_json', 'flag': '-d', 'is_path': True, 'pre_exist': True, 'required': True,
         'help': 'The json file containing model meta-data for documentation'},
        {'name': 'doc_schema_info', 'flag': '-s', 'is_path': True, 'pre_exist': True, 'required': False,
         'help': 'The schema for the model meta-data json'},

    ]
    chained_commands = []

    def validate_doc_schema(self, schema_path, docjson_path):
        """Validates docjson_path file with schema_path file
        Args:
            schema_path (str | os.PathLike): Schema path file
            docjson_path (str | os.PathLike): Documentation JSON path file
        Returns:
            docjson (Dict): Json data loaded as a dictionary
        """
        with open(schema_path, "r") as f:
            schema = json.load(f)

        with open(docjson_path, "r") as f:
            docjson = json.load(f)

        if "datasets" not in docjson:
            raise ValidationError(f"key \'datasets\' not found inside {docjson_path}")

        datasets = docjson["datasets"]
        for i, dataset in enumerate(datasets):
            try:
                validate(instance=dataset, schema=schema)
            except ValidationError as e:
                raise ValidationError(f"doc schema validation error for dataset idx {i}: {e.message}")

        return docjson, schema

    def json_to_mdtxt(self, json_data, full_schema, data_path, doc_out_dir):
        """Convert json data to markdown text with schemas provided
        Args:
            json_data (dict): Json data as dictionary
            full_schema (dict): Full schema file as dictionary
            data_path (str | os.PathLike): Path to data folder for any relative file paths
            doc_out_dir (str | os.PathLike): Path to documentation file output folder for any relative file paths
        """
        schema_id = full_schema["$id"]
        json_to_md_generator = DefaultJsonToMarkdownGenerator
        if schema_id == "https://docs.riskdatalibrary.org/en/0__2__0/rdls_schema.json":
            # RDLS v0.2
            json_to_md_generator = RDLS_0_2_0_JsonToMarkdownGenerator
        else:
            self.logger.warning(f"WARN: Unsupported formatting for following schema: {schema_id}. Using DefaultJsonToMarkdownGenerator output")
        gen = json_to_md_generator(full_schema, data_path, doc_out_dir)
        return gen.generate(json_data, generate_toc=True)

    def run(self):
        if not os.path.exists(self.doc_json):
            raise FileNotFoundError(f'Could not locate doc_json file: {self.doc_json}, Cannot generate documentation')
        if not self.doc_schema_info:
            self.doc_schema_info = resources.files('rdls').joinpath('rdls_schema.json')
            if not os.path.exists(self.doc_schema_info):
                raise FileNotFoundError(f'Could not locate doc_schema_info file: {self.doc_schema_info}, Cannot generate documentation')

        doc_out_dir = Path(self.doc_out_dir)
        doc_json = Path(self.doc_json)
        data_path = doc_json.parent
        doc_schema_info = Path(self.doc_schema_info)
        doc_file = Path(doc_out_dir, 'doc.md')
        json_data, schema = self.validate_doc_schema(doc_schema_info, doc_json)

        with open(doc_file, "w") as f:
            mdtxt = self.json_to_mdtxt(json_data, schema, data_path, doc_out_dir)
            f.write(mdtxt)




__all__ = [
    'GenerateDocumentation'
]

import os

from ..base import ComputationStep
from ..generate.doc import GenerateModelDocumentation


class GenerateDocumentation(ComputationStep):
    """
    Generate Documentation for model from the config file
    """

    # Override params
    step_params = []
    # Add params from each sub command not in 'step_params'
    chained_commands = [
        GenerateModelDocumentation,
    ]

    def run(self):
        #  setup documentation output dir
        if not self.doc_out_dir:
            self.doc_out_dir = GenerateModelDocumentation._get_output_dir(self)

        # create documentation output dir
        if not os.path.exists(self.doc_out_dir):
            os.makedirs(self.doc_out_dir)

        # generate Model Documentation
        self.kwargs['doc_out_dir'] = self.doc_out_dir
        GenerateModelDocumentation(**self.kwargs).run()

        # logger info
        self.logger.info('\nGenerate Documentation completed successfully in {}'.format(self.doc_out_dir))



import argparse
import logging

from .manager import bintoparquet, logger
from oasislmf.pytools.converters.data import SUPPORTED_BINTOPARQUET


def main():
    parser = argparse.ArgumentParser(description='Convert Binary to Parquet for various file types.',
                                     formatter_class=argparse.RawTextHelpFormatter)

    subparsers = parser.add_subparsers(dest='file_type', required=True, help='Type of file to convert')
    for file_type in SUPPORTED_BINTOPARQUET:
        parser_curr = subparsers.add_parser(file_type, help=f'bin to parquet tool for {file_type}')
        parser_curr.add_argument('-i', '--file_in', default="-", type=str, help='Input file path')
        parser_curr.add_argument('-o', '--file_out', default="-", type=str, help='Output file path')
        parser_curr.add_argument('-v', '--logging-level', type=int, default=30,
                                 help='logging level (debug:10, info:20, warning:30, error:40, critical:50)')
    args = parser.parse_args()
    kwargs = vars(args)

    file_type = kwargs.pop('file_type')
    file_in = kwargs.pop('file_in')
    file_out = kwargs.pop('file_out')

    # Set up logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    bintoparquet(file_in, file_out, file_type, **kwargs)


if __name__ == '__main__':
    main()



from oasislmf.pytools.aal.data import AAL_meanonly_headers, AAL_meanonly_dtype, AAL_meanonly_fmt


headers = AAL_meanonly_headers
dtype = AAL_meanonly_dtype
fmt = AAL_meanonly_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



__all__ = [
    "amplifications_tocsv",
    "cdf_tocsv",
    "complex_items_tocsv",
    "coverages_tocsv",
    "fm_tocsv",
    "footprint_tocsv",
    "gul_tocsv",
    "lossfactors_tocsv",
    "occurrence_tocsv",
    "vulnerability_tocsv",
]

from .amplifications import amplifications_tocsv
from .cdf import cdf_tocsv
from .complex_items import complex_items_tocsv
from .coverages import coverages_tocsv
from .fm import fm_tocsv
from .footprint import footprint_tocsv
from .gul import gul_tocsv
from .lossfactors import lossfactors_tocsv
from .occurrence import occurrence_tocsv
from .vulnerability import vulnerability_tocsv



from oasislmf.pytools.common.data import amplifications_headers, amplifications_dtype, amplifications_fmt


headers = amplifications_headers
dtype = amplifications_dtype
fmt = amplifications_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.elt.data import MELT_headers, MELT_dtype, MELT_fmt


headers = MELT_headers
dtype = MELT_dtype
fmt = MELT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



from oasislmf.pytools.common.data import fm_profile_step_headers, fm_profile_step_dtype, fm_profile_step_fmt


headers = fm_profile_step_headers
dtype = fm_profile_step_dtype
fmt = fm_profile_step_fmt
cli_support = ['bintocsv', 'csvtobin']



#!/usr/bin/env python

from contextlib import ExitStack
import logging

from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.data import TOOL_INFO
import pandas as pd

logger = logging.getLogger(__name__)


def default_tobin(stack, file_in, file_out, file_type):
    dtype = TOOL_INFO[file_type]["dtype"]

    file_in = resolve_file(file_in, "rb", stack)

    df = pd.read_parquet(file_in)
    data = df.to_records(index=False).astype(dtype)
    data.tofile(file_out)


def parquettobin(file_in, file_out, file_type, **kwargs):
    """Convert parquet file to bin file based on file type
    Args:
        file_in (str | os.PathLike): Input file path
        file_out (str | os.PathLike): Output file path
        file_type (str): File type str from SUPPORTED_PARQUETTOBIN
    """
    with ExitStack() as stack:
        file_out = resolve_file(file_out, "wb", stack)
        default_tobin(stack, file_in, file_out, file_type, **kwargs)



import json

from oasislmf.utils.documentation.jsontomd import BaseJsonToMarkdownGenerator


class DefaultJsonToMarkdownGenerator(BaseJsonToMarkdownGenerator):
    """
    Default JSON to Markdown Generator class.
    Naively iterates through the dict and outputs with limited formatting.
    """

    def json_array_to_mdtable(self, data, ref):
        array_schema = self._resolve_internal_ref(ref)
        array_keys = array_schema["properties"].keys()
        headers = []
        for k in array_keys:
            if "title" in array_schema["properties"][k]:
                headers.append(array_schema["properties"][k]["title"])
            else:
                headers.append(k)
        rows = []
        for entry in data:
            row = []
            for k in array_keys:
                v = entry.get(k, "")
                if isinstance(v, list):
                    rets = []
                    for v_ in v:
                        if isinstance(v_, dict):
                            v_ = json.dumps(v_, indent=4).replace('\n', '<br>').replace(' ', '&nbsp;')
                        rets.append(str(v_))
                    v = ",<br>".join(rets)
                elif isinstance(v, dict):
                    pretty_json = json.dumps(v, indent=4).replace('\n', '<br>').replace(' ', '&nbsp;')
                    v = f"<details><summary>Expand</summary>{pretty_json}</details>"
                else:
                    v = str(v)
                row.append(v)
            rows.append(row)
        self.md.add_table(headers, rows)

    def generate_data(self, data, properties_schema, header_level):
        for key, value in data.items():
            key_title = key
            schema = properties_schema.get(key, {})
            if "type" in schema:
                if schema["type"] == "array":
                    self.md.add_header(key_title, level=header_level)
                    items = schema.get("items", {})
                    if isinstance(items, dict) and "$ref" in items:
                        self.json_array_to_mdtable(value, items["$ref"])
                    else:
                        self.md.add_list(value)
                elif schema["type"] == "object":
                    self.md.add_header(key_title, level=header_level)
                    self.generate_data(value, schema.get("properties", {}), header_level + 1)
                else:
                    self.md.add_header(key_title, level=header_level)
                    self.md.add_text(value)
            elif "$ref" in schema:
                self.md.add_header(key_title, level=header_level)
                ref_schema = self._resolve_internal_ref(schema["$ref"])
                self.generate_data(value, ref_schema.get("properties", {}), header_level + 1)
            else:
                self.md.add_header(key_title, level=header_level)
                if isinstance(value, list):
                    for i, v in enumerate(value):
                        self.md.add_header(f"Item {i}", header_level + 1)
                        self.generate_data(v, properties_schema, header_level + 2)
                elif isinstance(value, dict):
                    for k, v in value.items():
                        self.md.add_header(k, level=header_level + 1)
                        self.generate_data(v, properties_schema, header_level + 2)
                else:
                    self.md.add_text(value)

    def generate(self, json_data, generate_toc=False):
        self.md.add_header("Documentation", level=1)
        self.generate_data(json_data, self.full_schema["properties"], header_level=3)
        return self.md.get_markdown(generate_toc=generate_toc)



import json
import os
import argparse
import shutil
from tqdm.auto import tqdm
from contextlib import ExitStack

import pandas as pd
import pyarrow as pa
from pyarrow import parquet as pq

from oasis_data_manager.filestore.backends.local import LocalStorage
from oasislmf.pytools.getmodel.footprint import Footprint


def convert_bin_to_parquet_event(static_path: str, **kwargs) -> None:
    """
    Converts the data from a binary file to a parquet file.

    Args:
        static_path: (str) the path to the static file

    Returns: None
    """
    root_path = os.path.join(static_path, 'footprint.parquet')
    try:
        shutil.rmtree(root_path)
    except FileNotFoundError:
        pass
    with ExitStack() as stack:
        storage = LocalStorage(
            root_dir=static_path,
            cache_dir=None,
        )
        footprint_obj = stack.enter_context(Footprint.load(storage,
                                                           ignore_file_type={'csv', 'parquet'}))
        index_data = footprint_obj.footprint_index

        meta_data = {
            "num_intensity_bins": footprint_obj.num_intensity_bins,
            "has_intensity_uncertainty": True if footprint_obj.has_intensity_uncertainty == 1 else False
        }

        for event_id in tqdm(index_data.keys(), desc="processing events"):
            data_slice = footprint_obj.get_event(event_id)
            df = pd.DataFrame(data_slice)
            df["event_id"] = event_id
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=root_path,
                partition_cols=['event_id'],
                compression="BROTLI"
            )
        with storage.open('footprint_parquet_meta.json', 'w') as outfile:
            json.dump(meta_data, outfile)


def convert_bin_to_parquet_chunk(static_path, chunk_size, **kwargs) -> None:
    """
    Converts the data from a binary file to a parquet file.

    Args:
        static_path: (str) the path to the static file
        chunk_size: target raw size of the partition

    Returns: None
    """
    root_path = os.path.join(static_path, 'footprint_chunk')
    try:
        shutil.rmtree(root_path)
    except FileNotFoundError:
        pass
    os.mkdir(root_path)
    density = (0.0, 0)
    with ExitStack() as stack:
        storage = LocalStorage(
            root_dir=static_path,
            cache_dir=None,
        )
        footprint_obj = stack.enter_context(
            Footprint.load(storage, ignore_file_type={'csv', 'parquet'})
        )
        index_data = footprint_obj.footprint_index
        meta_data = {
            "num_intensity_bins": footprint_obj.num_intensity_bins,
            "has_intensity_uncertainty": True
            if footprint_obj.has_intensity_uncertainty == 1
            else False
        }

        event_data = []
        for event_id in tqdm(index_data.keys(), desc="parsing index file"):
            data_slice = footprint_obj.get_event(event_id)
            df = pd.DataFrame(data_slice)
            min_areaperil_id = min(df['areaperil_id'])
            max_areaperil_id = max(df['areaperil_id'])
            event_data.append((min_areaperil_id, max_areaperil_id, event_id))
            if df.shape[0]:
                if max_areaperil_id == min_areaperil_id:
                    cur_density = 1
                else:
                    cur_density = df.shape[0] / (max_areaperil_id - min_areaperil_id)
                density = ((density[0] * density[1] + cur_density) / (density[1] + 1),
                           density[1] + 1)

        event_data.sort(key=lambda x: x[0])

        current_chunk = []
        current_size = 0
        count = 1
        footprint_lookup = []

        for min_apid, max_apid, event_id in tqdm(event_data, desc="processing chunks"):
            footprint_lookup.append({
                'event_id': event_id,
                'partition': count,
                'min_areaperil_id': min_apid,
                'max_areaperil_id': max_apid
            })  # size?

            data_slice = footprint_obj.get_event(event_id)
            df = pd.DataFrame(data_slice)
            df["event_id"] = event_id

            current_chunk.append(df)
            current_size += df.memory_usage(deep=True).sum()

            if (current_size < chunk_size):
                continue

            pq.write_table(
                pa.Table.from_pandas(
                    pd.concat(current_chunk, ignore_index=True)
                ),
                os.path.join(root_path, f'footprint_{count}.parquet'),
                compression="BROTLI"
            )

            current_chunk = []
            current_size = 0
            count += 1

        if current_chunk:
            pq.write_table(
                pa.Table.from_pandas(
                    pd.concat(current_chunk, ignore_index=True)
                ),
                os.path.join(root_path, f'footprint_{count}.parquet'),
                compression='BROTLI'
            )

        footprint_lookup_df = pd.DataFrame(footprint_lookup)

        footprint_lookup_df.to_parquet(f'{static_path}/footprint_lookup.parquet', index=False)
        # Writing to other file types min and max ids
        no_partition = footprint_lookup_df.drop(columns=['partition'])
        no_partition.to_pickle(f'{static_path}/footprint_lookup.bin')
        no_partition.to_csv(f'{static_path}/footprint_lookup.csv', index=False)

        with open(f'{static_path}/footprint_parquet_meta.json', 'w') as outfile:
            json.dump(meta_data, outfile)
    print(f"event area_peril_id density is {density[0]}")


default_chunk_size = 1024 * 1024 * 8

parser = argparse.ArgumentParser()
parser.add_argument('-p', '--static-path', help='path to the folder containing the footprint files', default='.')
parser.add_argument('--chunk', help='if flag is set will create chunk footprint with the target raw size of chunk_size', action='store_true')
parser.add_argument('--chunk-size', help=f'target size for chunk parquet, default {default_chunk_size / 1024 * 1024}MB',
                    type=int, default=default_chunk_size)


def main():
    kwargs = vars(parser.parse_args())
    if kwargs['static_path'] == '.':
        kwargs['static_path'] = str(os.getcwd())

    if kwargs['chunk']:
        convert_bin_to_parquet_chunk(**kwargs)
    else:
        convert_bin_to_parquet_event(**kwargs)


if __name__ == '__main__':
    main()



import socket
import threading
import json
import os
import time
import sys
from tqdm import tqdm
from oasislmf.utils.defaults import SERVER_UPDATE_TIME, SERVER_DEFAULT_PORT, SERVER_DEFAULT_IP
import logging


class GulProgressServer:
    def __init__(self, total, host=None, port=None):
        """
        Args:
            total (int): How many events the server has
            host (str, optional): Non-default host to use. Defaults to `OASIS_SOCKET_SERVER_IP` or SERVER_DEFAULT_IP if unset.
            port (int, optional): Non-default port to use. Defaults to `OASIS_SOCKET_SERVER_PORT` or SERVER_DEFAULT_PORT if unset.
        """
        self.host = os.environ.get('OASIS_SOCKET_SERVER_IP', SERVER_DEFAULT_IP) if host is None else host
        self.port = int(os.environ.get('OASIS_SOCKET_SERVER_PORT', SERVER_DEFAULT_PORT)) if port is None else port
        self.total = total
        self.counter = 0
        self.counter_lock = threading.Lock()
        self.running = False
        self._accept_thread = False

    def start(self):
        self.running = True
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        try:
            self.server_socket.bind((self.host, self.port))
        except OSError:  # Desired socket unavailable
            self.server_socket.bind((self.host, 0))
            self.port = self.server_socket.getsockname()[1]
            logging.info(f"Port for socket server changed to {self.port}")
            os.environ['OASIS_SOCKET_SERVER_PORT'] = str(self.port)

        self.server_socket.listen(5)

        self._accept_thread = threading.Thread(target=self._accept_loop, daemon=True)
        self._accept_thread.start()

    def _accept_loop(self):
        while self.running:
            try:
                client_socket, _ = self.server_socket.accept()
                threading.Thread(target=self._handle_client, args=(client_socket,), daemon=True).start()
            except OSError:
                break  # socket was closed

    def _handle_client(self, client_socket):
        data = self._read_all(client_socket)
        payload = json.loads(data)
        with self.counter_lock:
            if 'terminate' in payload:
                self.counter = self.total
                self.stop()
            self.counter += int(payload.get("events_complete", 0))

    def _read_all(self, client_socket):
        buffer = b""
        while not buffer.endswith(b"\n"):
            chunk = client_socket.recv(1024)
            if not chunk:  # client disconnected
                break
            buffer += chunk
        return buffer.decode("utf-8").strip()

    def stop(self):
        self.running = False
        self.server_socket.close()
        if self._accept_thread:
            self._accept_thread.join(timeout=1)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()


def main():
    if len(sys.argv) < 2:
        raise ValueError("Socket server must be called with an argument for pbar length")
    try:
        total = int(sys.argv[1])
    except Exception:
        raise TypeError("Socket server argument must be an integer")
    with (GulProgressServer(total) as server,
          tqdm(total=total, unit="events", desc="Gul events completed", leave=True) as pbar):
        counter = 0
        while counter < total:
            with server.counter_lock:
                if counter != server.counter:
                    pbar.update(server.counter - counter)
                    counter = server.counter
            time.sleep(SERVER_UPDATE_TIME)



from oasislmf.pytools.common.data import complex_items_meta_headers, complex_items_meta_dtype, complex_items_meta_fmt


headers = complex_items_meta_headers
dtype = complex_items_meta_dtype
fmt = complex_items_meta_fmt
cli_support = ['bintocsv', 'csvtobin']



import json
import websocket
import socket
import os
import logging
from oasislmf.utils.defaults import SERVER_DEFAULT_PORT, SERVER_DEFAULT_IP


def oasis_ping(data):
    """
    Sends a JSON message to either a websocket server or a socket server.

    If `analysis_pk` is in the data, `OASIS_WEBSOCKET_URL` and `OASIS_WEBSOCKET_URL` are in environment, sends a websocket message.
    If `analysis_pk` but missing variables, no message sent.
    Else, websocket sent to `OASIS_SOCKET_SERVER_IP` `OASIS_SOCKET_SERVER_PORT` defaulted to 127.0.0.1 8888.

    For a specific socket or websocket, use `oasis_ping_socket` or `oasis_ping_websocket` with the target location.

    Args:
        data (dict): dictionary of data: JSON serialisable

    Returns:
        Boolean: whether attempted call gets through
    """
    msg = json.dumps(data)
    if data.get('analysis_pk', None) is not None:
        if all(item in os.environ for item in ['OASIS_WEBSOCKET_URL', 'OASIS_WEBSOCKET_PORT']):
            return oasis_ping_websocket(f"{os.environ['OASIS_WEBSOCKET_URL']}:{os.environ['OASIS_WEBSOCKET_PORT']}/ws/analysis-status/", msg)
        logging.error("Missing environment variables `OASIS_WEBSOCKET_URL` and `OASIS_WEBSOCKET_PORT`.")
        return False
    target = (os.environ.get("OASIS_SOCKET_SERVER_IP", SERVER_DEFAULT_IP), int(os.environ.get("OASIS_SOCKET_SERVER_PORT", SERVER_DEFAULT_PORT)))
    return oasis_ping_socket(target, msg)


def oasis_ping_socket(target, data):
    """
    Sends a JSON message to a target socket

    Args:
        target ((str, int)): IP and port to hit
        data (str): JSON dumped string

    Returns:
        Boolean: whether attempted call gets through
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as oasis_socket:
            oasis_socket.connect(target)
            oasis_socket.sendall(data.encode('utf-8'))
        return True
    except ConnectionRefusedError as e:
        logging.error(f"oasis_ping_socket could not connect: {e}")
        return False


def oasis_ping_websocket(ws_url, data):
    """
    Sends a JSON message to a target websocket

    Args:
        ws_url (str): URL to hit (e.g. "ws://oasis-websocket:8001/ws/analysis-status/")
        data (str): JSON dumped string

    Returns:
        Boolean: whether attempted call gets through
    """
    ws = websocket.WebSocket()
    try:
        ws.connect(ws_url)
        ws.send(data)
        return True
    except Exception as e:
        logging.error(f"oasis_ping_websocket could not connect: {e}")
        return False
    finally:
        ws.close()



from oasislmf.pytools.common.data import vulnerability_weight_headers, vulnerability_weight_dtype, vulnerability_weight_fmt


headers = vulnerability_weight_headers
dtype = vulnerability_weight_dtype
fmt = vulnerability_weight_fmt
cli_support = ['bintocsv', 'csvtobin']



import numba as nb
import numpy as np

from oasislmf.pytools.lec.utils import get_wheatsheaf_items_idx_data


@nb.njit(cache=True, error_model="numpy")
def get_wheatsheaf_max_count(
    wheatsheaf_items,
    wheatsheaf_items_start_end,
    max_summary_id,
):
    maxcount = np.full((max_summary_id), -1, dtype=np.int64)
    for start, end in wheatsheaf_items_start_end:
        summary_id = wheatsheaf_items[start]["summary_id"]
        size = end - start
        if size < maxcount[summary_id - 1]:
            continue
        maxcount[summary_id - 1] = size
    return maxcount


@nb.njit(cache=True, error_model="numpy")
def fill_wheatsheaf_mean_items(
    wheatsheaf_mean_items,
    wheatsheaf_items,
    wheatsheaf_items_start_end,
    maxcounts,
    max_summary_id,
    num_sidxs,
):
    # Track start and end indices for each summary_id
    items_start_end = np.full((max_summary_id, 2), -1, dtype=np.int32)

    # Compute cumulative start indices
    pos = 0
    for idx in range(max_summary_id):
        if maxcounts[idx] > 0:
            items_start_end[idx][0] = pos  # Start index
            pos += maxcounts[idx]
            items_start_end[idx][1] = pos  # End index

    for idx in range(max_summary_id * num_sidxs):
        ws_start, ws_end = wheatsheaf_items_start_end[idx]
        if ws_start == -1:
            continue
        sidx, summary_id = get_wheatsheaf_items_idx_data(idx, num_sidxs)
        filtered_items = wheatsheaf_items[ws_start:ws_end]
        sorted_idxs = np.argsort(filtered_items["value"])[::-1]
        sorted_items = filtered_items[sorted_idxs]

        wsm_start, wsm_end = items_start_end[summary_id - 1]

        for i, item in enumerate(sorted_items):
            # Compute position in the flat array
            insert_idx = wsm_start + i

            # Store values
            wheatsheaf_mean_items[insert_idx]["summary_id"] = summary_id
            wheatsheaf_mean_items[insert_idx]["value"] += item["value"]

    return items_start_end



# join-summary-info/manager.py

import logging
import numpy as np
from pathlib import Path
import pyarrow.parquet as pq
import pyarrow as pa
import tempfile
from contextlib import ExitStack

from oasislmf.pytools.utils import redirect_logging


logger = logging.getLogger(__name__)


def load_summary_info(stack, summaryinfo_file):
    """Load summary-info data into an array as strings to maintain sigfigs/formatting
    Args:
        stack (ExitStack): Exit Stack
        summaryinfo_file (str | os.PathLike): Path to summary-info csv file
    Returns:
        full_summary_data (ndarray[object]): Array of strings, indexed by Summary Id
        headers (List[str]): List of strings for summary info headers to add to data
        max_summary_id (int): Max Summary ID
    """
    summaryinfo_file = Path(summaryinfo_file)

    if summaryinfo_file.suffix == ".csv":
        summary_ids = []
        summary_data = []

        with stack.enter_context(open(summaryinfo_file, "r")) as fin:
            headers = fin.readline().strip().split(",")
            summary_id_col_idx = headers.index("summary_id")

            for line in fin:
                row = line.strip().split(",")
                summary_ids.append(row[summary_id_col_idx])
                summary_data.append(",".join(row[:summary_id_col_idx] + row[summary_id_col_idx + 1:]))

        headers = headers[:summary_id_col_idx] + headers[summary_id_col_idx + 1:]

        summary_ids = np.array(summary_ids, dtype=np.int64)
        summary_data = np.array(summary_data, dtype=object)
    elif summaryinfo_file.suffix == ".parquet":
        table = pq.read_table(summaryinfo_file)
        df = table.to_pandas()

        if "summary_id" not in df.columns:
            raise ValueError("Missing 'summary_id' column in summary info file.")

        summary_ids = df["summary_id"].to_numpy(dtype=np.int64)
        summary_data = df.drop(columns=["summary_id"]).astype(str).agg(",".join, axis=1).to_numpy(dtype=object)
        headers = [col for col in df.columns if col != "summary_id"]
    else:
        raise ValueError(f"Unsupported file format {summaryinfo_file.suffix}.")

    max_summary_id = summary_ids.max()
    full_summary_data = np.full((max_summary_id + 1,), "," * (len(headers) - 1), dtype=object)
    for i in range(len(summary_ids)):
        full_summary_data[summary_ids[i]] = summary_data[i]

    return full_summary_data, headers, max_summary_id


def run(
    summaryinfo_file,
    data_file,
    output_file,
):
    """Join the summary-info file to the given ORD data file based on SummaryId
    Args:
        summaryinfo_file (str | os.PathLike): Path to summary-info file
        data_file (str | os.PathLike): Path to ORD output data file (e.g. SELT, MPLT, AAL, PSEPT) 
        output_file (str | os.PathLike): Path to combined output file
    """
    summaryinfo_file = Path(summaryinfo_file)
    data_file = Path(data_file)
    output_file = Path(output_file)

    valid_file_suffix = [".csv", ".parquet"]

    if summaryinfo_file.suffix not in valid_file_suffix or \
       data_file.suffix not in valid_file_suffix or \
       output_file.suffix not in valid_file_suffix:
        raise ValueError(f"All files must be in one of {valid_file_suffix} formats.")

    if not (summaryinfo_file.suffix == data_file.suffix == output_file.suffix):
        raise ValueError("Summary info, data, and output files must all have the same format.")

    with ExitStack() as stack:
        summary_data, summary_headers, max_summary_id = load_summary_info(stack, summaryinfo_file)

        # Use a temporary file if input and output are the same
        same_file = data_file.resolve() == output_file.resolve()
        temp_output = None

        if same_file:
            temp_output = tempfile.NamedTemporaryFile(mode="w", delete=False)
            temp_output_file = Path(temp_output.name)
        else:
            temp_output_file = output_file

        if data_file.suffix == ".csv":
            with stack.enter_context(open(data_file, "r")) as data_fin, stack.enter_context(open(temp_output_file, "w")) as fout:
                data_headers = data_fin.readline().strip().split(",")
                fout.write(",".join(data_headers + summary_headers) + "\n")

                summary_id_col_idx = data_headers.index("SummaryId")
                for line in data_fin:
                    row = line.strip().split(",")
                    summary_id = int(row[summary_id_col_idx])
                    if summary_id > max_summary_id:
                        fout.write(line.strip() + ("," * len(summary_headers)) + "\n")
                    else:
                        fout.write(line.strip() + "," + summary_data[summary_id] + "\n")
        elif data_file.suffix == ".parquet":
            table = pq.read_table(data_file)
            df = table.to_pandas()

            if "SummaryId" not in df.columns:
                raise ValueError("Missing 'SummaryId' column in data file.")

            df[summary_headers] = df["SummaryId"].apply(
                lambda sid: summary_data[sid] if sid <= max_summary_id else ""
            ).str.split(",", expand=True)
            pq.write_table(pa.Table.from_pandas(df), temp_output_file)
        else:
            raise ValueError(f"Unsupported file format {data_file.suffix}.")

        # Replace the original file if needed
        if same_file and temp_output:
            temp_output.close()
            temp_output_file.replace(output_file)
            temp_output_file.unlink(missing_ok=True)


@redirect_logging(exec_name='join-summary-info')
def main(summaryinfo=None, data=None, output=None, **kwargs):
    run(
        summaryinfo_file=summaryinfo,
        data_file=data,
        output_file=output,
    )



"""
This file contains specific functionality to read and process items files.
"""
import logging
import os

import numpy as np
import numba as nb
from numba.typed import Dict, List
from numba.types import int32 as nb_int32
from numba.types import int8 as nb_int8

from oasislmf.pytools.common.data import nb_areaperil_int, nb_oasis_float, nb_oasis_int, items_dtype
from oasislmf.pytools.gul.utils import append_to_dict_value
from oasislmf.pytools.gulmc.aggregate import gen_empty_areaperil_vuln_idx_to_weights
from oasislmf.pytools.gulmc.common import ITEM_MAP_KEY_TYPE, ITEM_MAP_VALUE_TYPE


logger = logging.getLogger(__name__)


def read_items(input_path, ignore_file_type=set(), dynamic_footprint=False, legacy=False):
    """Load the items from the items file.

    Args:
        input_path (str): the path pointing to the file
        ignore_file_type (Set[str]): file extension to ignore when loading.

    Returns:
        Tuple[Dict[int, int], List[int], Dict[int, int], List[Tuple[int, int]], List[int]]
          vulnerability dictionary, vulnerability IDs, areaperil to vulnerability index dictionary,
          areaperil ID to vulnerability index array, areaperil ID to vulnerability array
    """
    input_files = set(os.listdir(input_path))

    if "items.bin" in input_files and "bin" not in ignore_file_type:
        items_fname = os.path.join(input_path, 'items.bin')
        logger.debug(f"loading {items_fname}")
        items = np.memmap(items_fname, dtype=items_dtype, mode='r')

    elif "items.csv" in input_files and "csv" not in ignore_file_type:
        items_fname = os.path.join(input_path, 'items.csv')
        logger.debug(f"loading {items_fname}")
        items = np.loadtxt(items_fname, dtype=items_dtype, delimiter=",", skiprows=1, ndmin=1)

    else:
        raise FileNotFoundError(f'items file not found at {input_path}')

    return items


@nb.njit(cache=True, fastmath=True)
def generate_item_map(items, coverages, valid_areaperil_id, agg_vuln_to_vulns):
    """Generate item_map; requires items to be sorted.

    Args:
        items (numpy.ndarray[int32, int32, int32]): 1-d structured array storing
            `item_id`, `coverage_id`, `group_id` for all items.
            items need to be sorted by increasing areaperil_id, vulnerability_id
            in order to output the items in correct order.
        coverages (numpy.ndarray): coverage id to information on items
        valid_areaperil_id (numpy.ndarray[int32]): list of non-filtered area_peril_id (None is no filter)
        agg_vuln_to_vulns (dict[int, list[int]]): map of aggregate vulnerability id to list of vulnerability ids.

    Returns:
        item_map (Dict[ITEM_MAP_KEY_TYPE, ITEM_MAP_VALUE_TYPE]): dict storing
            the mapping between areaperil_id, vulnerability_id to item.
        areaperil_ids_map (Dict[int, Dict[int, int]]) dict storing the mapping between each
            areaperil_id and all the vulnerability ids associated with it.
        vuln id to vuln idx for each vulnerability in each areaperil, list of all used vulnerability ids.
        agg_vuln_to_vuln_idxs dict[int, list[int]]: map between aggregate vulnerability id and the list of indices where the individual vulnerability_ids
          that compose it are stored in `vuln_array`.
        areaperil_vuln_idx_to_weight dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]: map between the areaperil id and the index where the vulnerability function
          is stored in `vuln_array` and the vulnerability weight.

    """
    item_map = Dict.empty(ITEM_MAP_KEY_TYPE, List.empty_list(ITEM_MAP_VALUE_TYPE))
    areaperil_ids_map = Dict.empty(nb_areaperil_int, Dict.empty(nb_int32, nb_int8))
    vuln_dict = Dict()
    vuln_idx = 0
    areaperil_vuln_idx_to_weight = gen_empty_areaperil_vuln_idx_to_weights()

    agg_vuln_to_vuln_idxs = Dict.empty(nb_int32, List.empty_list(nb_int32))

    for j, item in enumerate(items):
        areaperil_id = item['areaperil_id']
        vulnerability_id = item['vulnerability_id']

        if valid_areaperil_id is not None and item['areaperil_id'] not in valid_areaperil_id:
            continue

        append_to_dict_value(item_map, tuple((areaperil_id, vulnerability_id)), j, ITEM_MAP_VALUE_TYPE)
        coverages[item['coverage_id']]['max_items'] += 1

        # Populate vuln_dict, to map all used vuln_id to vuln_i
        if item['vulnerability_id'] not in vuln_dict:
            if item['vulnerability_id'] in agg_vuln_to_vulns:  # vulnerability is an aggregate
                for sub_vuln_id in agg_vuln_to_vulns[item['vulnerability_id']]:
                    if sub_vuln_id not in vuln_dict:
                        vuln_dict[sub_vuln_id] = nb_oasis_int(vuln_idx)
                        vuln_idx += 1
                agg_vuln_to_vuln_idxs[item['vulnerability_id']] = List([vuln_dict[vuln] for vuln in agg_vuln_to_vulns[item['vulnerability_id']]])

            else:  # single vulnerability
                vuln_dict[item['vulnerability_id']] = nb_oasis_int(vuln_idx)
                vuln_idx += 1

        if item['vulnerability_id'] in agg_vuln_to_vulns:
            for sub_vuln_id in agg_vuln_to_vulns[item['vulnerability_id']]:
                areaperil_vuln_idx_to_weight[(nb_areaperil_int(areaperil_id), vuln_dict[sub_vuln_id])] = nb_oasis_float(0)
        else:
            item['vulnerability_idx'] = vuln_dict[item['vulnerability_id']]

        if areaperil_id not in areaperil_ids_map:
            areaperil_ids_map[areaperil_id] = Dict.empty(nb_int32, nb_int8)
        areaperil_ids_map[areaperil_id][vulnerability_id] = 0

    return item_map, areaperil_ids_map, vuln_dict, agg_vuln_to_vuln_idxs, areaperil_vuln_idx_to_weight



import logging
from pathlib import Path
import numpy as np
import pandas as pd
import pyarrow as pa

from oasislmf.pytools.common.data import oasis_float, write_ndarray_to_fmt_csv
from oasislmf.pytools.lec.aggreports.outputs.full_uncertainty import output_full_uncertainty
from oasislmf.pytools.lec.aggreports.outputs.mean_damage_ratio import output_mean_damage_ratio
from oasislmf.pytools.lec.aggreports.outputs.sample_mean import output_sample_mean, reorder_losses_by_summary_and_period
from oasislmf.pytools.lec.aggreports.outputs.wheatsheaf import fill_wheatsheaf_items
from oasislmf.pytools.lec.aggreports.outputs.wheatsheaf_mean import fill_wheatsheaf_mean_items, get_wheatsheaf_max_count
from oasislmf.pytools.lec.aggreports.write_tables import write_ept, write_ept_weighted, write_psept, write_psept_weighted, write_wheatsheaf_mean
from oasislmf.pytools.lec.data import FULL, MEANDR, MEANSAMPLE, PERSAMPLEMEAN
from oasislmf.pytools.lec.data import LOSSVEC2MAP_dtype, MEANMAP_dtype, WHEATKEYITEMS_dtype

logger = logging.getLogger(__name__)


class AggReports():
    def __init__(
        self,
        outmap,
        outloss_mean,
        outloss_sample,
        period_weights,
        max_summary_id,
        sample_size,
        no_of_periods,
        num_sidxs,
        use_return_period,
        returnperiods,
        lec_files_folder,
        output_binary,
        output_parquet,
    ):
        self.outmap = outmap
        self.outloss_mean = outloss_mean
        self.outloss_sample = outloss_sample
        self.period_weights = period_weights
        self.max_summary_id = max_summary_id
        self.sample_size = sample_size
        self.no_of_periods = no_of_periods
        self.num_sidxs = num_sidxs
        self.use_return_period = use_return_period
        self.returnperiods = returnperiods
        self.lec_files_folder = lec_files_folder
        self.output_binary = output_binary
        self.output_parquet = output_parquet

    def output_data(self, data, out_type):
        if self.output_binary:
            data.tofile(self.outmap[out_type]["file"])
        elif self.output_parquet:
            data_df = pd.DataFrame(data)
            data_table = pa.Table.from_pandas(data_df)
            self.outmap[out_type]["file"].write_table(data_table)
        else:
            write_ndarray_to_fmt_csv(
                self.outmap[out_type]["file"],
                data,
                self.outmap[out_type]["headers"],
                self.outmap[out_type]["fmt"]
            )

    def output_mean_damage_ratio(self, eptype, eptype_tvar, outloss_type):
        """Output Mean Damage Ratio
        Mean Damage Losses - This means do the loss calculation for a year using the event mean
        damage loss computed by numerical integration of the effective damageability distributions.
        Args:
            eptype (int): Exceedance Probability Type
            eptype_tvar (int): Exceedance Probability Type (Tail Value at Risk)
            outloss_type (string): Which loss to output
        """
        epcalc = MEANDR

        # Get row indices that are used
        row_used_indices = np.where(self.outloss_mean["row_used"])[0]

        # Allocate storage for the flat data array
        items_fp = Path(self.lec_files_folder, f"lec_mean_damage_ratio-{outloss_type}-items.bdat")
        items = np.memmap(items_fp, dtype=LOSSVEC2MAP_dtype, mode="w+", shape=(len(row_used_indices),))
        # Track start and end indices for each summary_id
        items_start_end = np.full((self.max_summary_id, 2), -1, dtype=np.int32)

        # Select the correct outloss values based on type
        # Required if-else condition as njit cannot resolve outloss_type inside []
        if outloss_type == "agg_out_loss":
            outloss_vals = self.outloss_mean["agg_out_loss"]
        elif outloss_type == "max_out_loss":
            outloss_vals = self.outloss_mean["max_out_loss"]
        else:
            raise ValueError(f"Error: Unknown outloss_type: {outloss_type}")

        # Populate items and items_start_end
        has_weights, used_period_no = output_mean_damage_ratio(
            items,
            items_start_end,
            row_used_indices,
            outloss_vals,
            self.period_weights,
            self.max_summary_id,
        )
        unused_period_weights = self.period_weights[~used_period_no]

        if has_weights:
            gen = write_ept_weighted(
                items,
                items_start_end,
                self.sample_size,
                epcalc,
                eptype,
                eptype_tvar,
                unused_period_weights,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id
            )
        else:
            gen = write_ept(
                items,
                items_start_end,
                self.no_of_periods,
                epcalc,
                eptype,
                eptype_tvar,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id

            )

        for data in gen:
            self.output_data(data, "ept")

    def output_full_uncertainty(self, eptype, eptype_tvar, outloss_type):
        """Output Full Uncertainty
        Full Uncertainty – this means do the calculation across all samples (treating the samples
        effectively as repeat years) - this is the most accurate of all the single EP Curves.
        Args:
            eptype (int): Exceedance Probability Type
            eptype_tvar (int): Exceedance Probability Type (Tail Value at Risk)
            outloss_type (string): Which loss to output
        """
        epcalc = FULL

        # Get row indices that are used
        row_used_indices = np.where(self.outloss_sample["row_used"])[0]

        # Allocate storage for the flat data array
        items_fp = Path(self.lec_files_folder, f"lec_full_uncertainty-{outloss_type}-items.bdat")
        items = np.memmap(items_fp, dtype=LOSSVEC2MAP_dtype, mode="w+", shape=(len(row_used_indices),))
        # Track start and end indices for each summary_id
        items_start_end = np.full((self.max_summary_id, 2), -1, dtype=np.int32)

        # Select the correct outloss values based on type
        # Required if-else condition as njit cannot resolve outloss_type inside []
        if outloss_type == "agg_out_loss":
            outloss_vals = self.outloss_sample["agg_out_loss"]
        elif outloss_type == "max_out_loss":
            outloss_vals = self.outloss_sample["max_out_loss"]
        else:
            raise ValueError(f"Error: Unknown outloss_type: {outloss_type}")

        # Populate items and items_start_end
        has_weights, used_period_no = output_full_uncertainty(
            items,
            items_start_end,
            row_used_indices,
            outloss_vals,
            self.period_weights,
            self.max_summary_id,
            self.num_sidxs,
        )
        unused_period_weights = self.period_weights[~used_period_no]

        if has_weights:
            gen = write_ept_weighted(
                items,
                items_start_end,
                1,
                epcalc,
                eptype,
                eptype_tvar,
                unused_period_weights,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id
            )
        else:
            gen = write_ept(
                items,
                items_start_end,
                self.no_of_periods * self.sample_size,
                epcalc,
                eptype,
                eptype_tvar,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id
            )

        for data in gen:
            self.output_data(data, "ept")

    def output_wheatsheaf_and_wheatsheafmean(self, eptype, eptype_tvar, outloss_type, output_wheatsheaf, output_wheatsheaf_mean):
        """Output Wheatsheaf and Wheatsheaf Mean
        Wheatsheaf, Per Sample EPT (PSEPT) – this means calculate the EP Curve for each sample and
        leave it at the sample level of detail, resulting in multiple “curves”.
        Wheatsheaf Mean, Per Sample mean EPT – this means average the loss at each return period of
        the Per Sample EPT.
        Args:
            eptype (int): Exceedance Probability Type
            eptype_tvar (int): Exceedance Probability Type (Tail Value at Risk)
            outloss_type (string): Which loss to output
            output_wheatsheaf (bool): Bool to Output Wheatsheaf
            output_wheatsheaf_mean (bool): Bool to Output Wheatsheaf Mean
        """
        epcalc = PERSAMPLEMEAN

        # Get row indices that are used
        row_used_indices = np.where(self.outloss_sample["row_used"])[0]

        wheatsheaf_items_file = Path(self.lec_files_folder, f"lec_wheatsheaf-items-{outloss_type}.bdat")
        wheatsheaf_items = np.memmap(
            wheatsheaf_items_file,
            dtype=WHEATKEYITEMS_dtype,
            mode="w+",
            shape=(len(row_used_indices)),
        )
        # Track start and end indices for each summary_id and sidx
        wheatsheaf_items_start_end = np.full((self.max_summary_id * self.num_sidxs, 2), -1, dtype=np.int32)

        # Select the correct outloss values based on type
        # Required if-else condition as njit cannot resolve outloss_type inside []
        if outloss_type == "agg_out_loss":
            outloss_vals = self.outloss_sample["agg_out_loss"]
        elif outloss_type == "max_out_loss":
            outloss_vals = self.outloss_sample["max_out_loss"]
        else:
            raise ValueError(f"Error: Unknown outloss_type: {outloss_type}")

        # Populate wheatsheaf_items and wheatsheaf_items_start_end
        has_weights, used_period_no = fill_wheatsheaf_items(
            wheatsheaf_items,
            wheatsheaf_items_start_end,
            row_used_indices,
            outloss_vals,
            self.period_weights,
            self.max_summary_id,
            self.num_sidxs,
        )
        unused_period_weights = self.period_weights[~used_period_no]

        if has_weights:
            mean_map = None

            if output_wheatsheaf_mean:
                mean_map_file = Path(self.lec_files_folder, f"lec_wheatsheaf_mean-map-{outloss_type}.bdat")
                mean_map = np.memmap(
                    mean_map_file,
                    dtype=MEANMAP_dtype,
                    mode="w+",
                    shape=(self.max_summary_id, len(self.returnperiods)),
                )

            if output_wheatsheaf:
                gen = write_psept_weighted(
                    wheatsheaf_items,
                    wheatsheaf_items_start_end,
                    self.no_of_periods,
                    eptype,
                    eptype_tvar,
                    unused_period_weights,
                    self.use_return_period,
                    self.returnperiods,
                    self.max_summary_id,
                    self.num_sidxs,
                    self.sample_size,
                    mean_map=mean_map,
                )
                for data in gen:
                    self.output_data(data, "psept")

            if output_wheatsheaf_mean:
                gen = write_wheatsheaf_mean(
                    mean_map,
                    eptype,
                    epcalc,
                    self.max_summary_id,
                )
                for data in gen:
                    self.output_data(data, "ept")
        else:
            if output_wheatsheaf:
                gen = write_psept(
                    wheatsheaf_items,
                    wheatsheaf_items_start_end,
                    self.no_of_periods,
                    eptype,
                    eptype_tvar,
                    self.use_return_period,
                    self.returnperiods,
                    self.max_summary_id,
                    self.num_sidxs,
                )
                for data in gen:
                    self.output_data(data, "psept")

            if not output_wheatsheaf_mean:
                return

            maxcounts = get_wheatsheaf_max_count(
                wheatsheaf_items,
                wheatsheaf_items_start_end,
                self.max_summary_id,
            )

            wheatsheaf_mean_items_file = Path(self.lec_files_folder, f"lec_wheatsheaf_mean-items-{outloss_type}.bdat")
            wheatsheaf_mean_items = np.memmap(
                wheatsheaf_mean_items_file,
                dtype=LOSSVEC2MAP_dtype,
                mode="w+",
                shape=(np.sum(maxcounts[maxcounts != -1])),
            )

            wheatsheaf_mean_items_start_end = fill_wheatsheaf_mean_items(
                wheatsheaf_mean_items,
                wheatsheaf_items,
                wheatsheaf_items_start_end,
                maxcounts,
                self.max_summary_id,
                self.num_sidxs,
            )

            gen = write_ept(
                wheatsheaf_mean_items,
                wheatsheaf_mean_items_start_end,
                self.no_of_periods,
                epcalc,
                eptype,
                eptype_tvar,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id,
                sample_size=self.sample_size
            )

            for data in gen:
                self.output_data(data, "ept")

    def output_sample_mean(self, eptype, eptype_tvar, outloss_type):
        """Output Sample Mean
        Sample Mean Losses – this means do the loss calculation for a year using the statistical
        sample event mean.
        Args:
            eptype (int): Exceedance Probability Type
            eptype_tvar (int): Exceedance Probability Type (Tail Value at Risk)
            outloss_type (string): Which loss to output
        """
        if self.sample_size == 0:
            logger.warning("aggreports.output_sample_mean, self.sample_size is 0, not outputting any sample mean")
            return
        epcalc = MEANSAMPLE

        # outloss_sample has all SIDXs plus -2 and -3
        reordered_outlosses_file = Path(self.lec_files_folder, f"lec_sample_mean-reordered_outlosses-{outloss_type}.bdat")
        reordered_outlosses = np.memmap(
            reordered_outlosses_file,
            dtype=np.dtype([
                ("row_used", np.bool_),
                ("value", oasis_float),
            ]),
            mode="w+",
            shape=(self.no_of_periods * self.max_summary_id),
        )

        # Select the correct outloss values based on type
        # Required if-else condition as njit cannot resolve outloss_type inside []
        if outloss_type == "agg_out_loss":
            outloss_vals = self.outloss_sample["agg_out_loss"]
        elif outloss_type == "max_out_loss":
            outloss_vals = self.outloss_sample["max_out_loss"]
        else:
            raise ValueError(f"Error: Unknown outloss_type: {outloss_type}")

        # Get row indices that are used
        row_used_indices = np.where(self.outloss_sample["row_used"])[0]

        # Reorder outlosses by summary_id and period_no
        reorder_losses_by_summary_and_period(
            reordered_outlosses,
            row_used_indices,
            outloss_vals,
            self.max_summary_id,
            self.no_of_periods,
            self.num_sidxs,
            self.sample_size,
        )

        # Get row indices that are used
        row_used_indices = np.where(reordered_outlosses["row_used"])[0]

        # Allocate storage for the flat data array
        items_fp = Path(self.lec_files_folder, f"lec_sample_mean-{outloss_type}-items.bdat")
        items = np.memmap(items_fp, dtype=LOSSVEC2MAP_dtype, mode="w+", shape=(len(row_used_indices),))
        # Track start and end indices for each summary_id
        items_start_end = np.full((self.max_summary_id, 2), -1, dtype=np.int32)

        # Populate items and items_start_end
        has_weights, used_period_no = output_sample_mean(
            items,
            items_start_end,
            row_used_indices,
            reordered_outlosses["value"],
            self.period_weights,
            self.max_summary_id,
            self.no_of_periods,
        )
        unused_period_weights = self.period_weights[~used_period_no]

        if has_weights:
            gen = write_ept_weighted(
                items,
                items_start_end,
                self.sample_size,
                epcalc,
                eptype,
                eptype_tvar,
                unused_period_weights,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id
            )
        else:
            gen = write_ept(
                items,
                items_start_end,
                self.no_of_periods,
                epcalc,
                eptype,
                eptype_tvar,
                self.use_return_period,
                self.returnperiods,
                self.max_summary_id
            )

        for data in gen:
            self.output_data(data, "ept")



import numba as nb
import numpy as np


"""
Custom numba heap implementation for for dtype np.int32

Stores num_compare + 2 ints, and sorts lexographically on all ints.
Last 2 ints are file_idx and row_number in file[file_idx]
"""


@nb.njit(cache=True, error_model="numpy")
def _resize_heap(heap, current_capacity):
    """Doubles the heap capacity"""
    new_capacity = current_capacity * 2
    new_heap = np.zeros((new_capacity, heap.shape[1]), dtype=heap.dtype)
    for i in range(current_capacity):
        new_heap[i] = heap[i]
    return new_heap


@nb.njit(cache=True, error_model="numpy")
def _lex_compare(a, b, row_size):
    """Lexicographically compare two rows from heap"""
    for i in range(row_size):
        if a[i] < b[i]:
            return True
        elif a[i] > b[i]:
            return False
    return False


@nb.njit(cache=True, error_model="numpy")
def _swap_rows(heap, i, j):
    """Swap rows i and j in heap"""
    for k in range(heap.shape[1]):
        temp = heap[i, k]
        heap[i, k] = heap[j, k]
        heap[j, k] = temp


@nb.njit(cache=True, error_model="numpy")
def _sift_down(heap, startpos, pos):
    """Heapq _siftdown"""
    # Follow the path to the root, moving parents down until finding a place
    # newitem fits.
    while pos > startpos:
        parentpos = (pos - 1) >> 1
        if _lex_compare(heap[pos], heap[parentpos], heap.shape[1]):
            _swap_rows(heap, pos, parentpos)
            pos = parentpos
            continue
        break


@nb.njit(cache=True, error_model="numpy")
def _sift_up(heap, pos, endpos):
    """Heapq _siftup"""
    startpos = pos
    # Bubble up the smaller child until hitting a leaf.
    childpos = 2 * pos + 1  # leftmost child position
    while childpos < endpos:
        # Set childpos to index of smaller child.
        rightpos = childpos + 1
        if rightpos < endpos and not _lex_compare(heap[childpos], heap[rightpos], heap.shape[1]):
            childpos = rightpos
        # Move the smaller child up.
        if pos != childpos:
            _swap_rows(heap, pos, childpos)
        pos = childpos
        childpos = 2 * pos + 1
    # The leaf at pos is empty now.  Put newitem there, and bubble it up
    # to its final resting place (by sifting its parents down).
    _sift_down(heap, startpos, pos)


@nb.njit(cache=True, error_model="numpy")
def heap_push(heap, size, element):
    """Heapq heappush"""
    if size >= len(heap):
        heap = _resize_heap(heap, len(heap))
    heap[size] = element
    _sift_down(heap, 0, size)
    return heap, size + 1


@nb.njit(cache=True, error_model="numpy")
def heap_pop(heap, size):
    """Heapq heappop"""
    if size <= 0:
        raise ValueError("Heap underflow: Cannot pop from an empty heap.")
    lastelt = heap[size - 1].copy()
    if size - 1 > 0:
        returnitem = heap[0].copy()
        heap[0] = lastelt
        _sift_up(heap, 0, size - 1)
        return returnitem, heap, size - 1
    return lastelt, heap, size - 1


@nb.njit(cache=True, error_model="numpy")
def init_heap(num_rows=4, num_compare=1):
    """Initialise heap with (num_compare + 2) * num_rows elements, where
    num_compare is the number of elements to order by in lexicographical
    order, and the remaining two elements are the file and row idxs

    Currently limited to dtype np.int32 as it is tricky to use custom dtypes
    """
    return np.zeros((num_rows, num_compare + 2), dtype=np.int32)



from pathlib import Path
import numpy as np
from oasislmf.pytools.pla.common import amp_factor_dtype
from oasislmf.pytools.pla.structure import read_lossfactors


def lossfactors_tobin(stack, file_in, file_out, file_type):
    if str(file_in) == "-":
        plafactors = read_lossfactors(
            ignore_file_type=set(["bin"]),
            use_stdin=True
        )
    else:
        lossfactors_fp = Path(file_in)
        plafactors = read_lossfactors(
            run_dir=lossfactors_fp.parent,
            ignore_file_type=set(["bin"]),
            filename=lossfactors_fp.name
        )

    # Write the 4-byte zero header
    np.array([0], dtype="i4").tofile(file_out)

    current_event_id = -1
    counter = 0
    factors = []
    for k, v in plafactors.items():
        if k[0] != current_event_id:
            if current_event_id != -1:
                np.array([current_event_id], dtype=np.int32).tofile(file_out)
                np.array([counter], dtype=np.int32).tofile(file_out)
                for af in factors:
                    np.array(af, dtype=amp_factor_dtype).tofile(file_out)
            current_event_id = k[0]
            counter = 0
            factors = []
        factors.append((k[1], v))
        counter += 1
    if current_event_id != -1:
        np.array([current_event_id], dtype=np.int32).tofile(file_out)
        np.array([counter], dtype=np.int32).tofile(file_out)
        for af in factors:
            np.array(af, dtype=amp_factor_dtype).tofile(file_out)



import numpy as np
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.gulmc.common import VALID_DAMAGE_TYPE
from oasislmf.utils.exceptions import OasisException


def _validate(data):
    from oasislmf.pytools.converters.csvtobin.manager import logger
    # Check first and last bin
    first_bin_index = data[0]["bin_index"]
    first_bin_from = data[0]["bin_from"]
    if first_bin_from != 0:
        logger.warning(f"Warning: Lower limit of first bin is not 0. Are you sure this is what you want?: bin_from={first_bin_from}")
    if first_bin_index != 1:
        raise OasisException(f"Error: First bin index must be 1, bin_index found: {first_bin_index}")

    last_bin_to = data[-1]["bin_to"]
    if last_bin_to != 1:
        logger.warning(f"Warning: Upper limit of last bin is not 1. Are you sure this is what you want?: bin_to={last_bin_to}")

    # Check contiguous bin_index
    indices = data["bin_index"]
    diffs = np.diff(indices)
    if not np.all(diffs == 1):
        idx = np.where(diffs != 1)[0][0]
        v1, v2 = indices[idx], indices[idx + 1]
        raise OasisException(f"Error: Non-contiguous bin_indices found in csv. ({v1}, {v2})")

    # Check interpolation damage values within range
    bin_froms = data["bin_from"]
    bin_tos = data["bin_to"]
    interpolations = data["interpolation"]
    mask_low = interpolations < bin_froms
    mask_high = interpolations > bin_tos
    mask_invalid = mask_low | mask_high
    if np.any(mask_invalid):
        bad_rows = np.where(mask_invalid)[0]
        error_msg = "\n".join(
            f"\tRow {i}: interpolation={interpolations[i]}, bin_from={bin_froms[i]}, bin_to={bin_tos[i]}"
            for i in bad_rows
        )
        raise OasisException(f"Error: Interpolation damage value outside of range.\n {error_msg}")

    # Check valid damage_type
    damages = data['damage_type']
    invalid_mask = ~np.isin(damages, list(VALID_DAMAGE_TYPE))
    invalid_indices = np.where(invalid_mask)[0]
    if invalid_indices.size > 0:
        invalid_values = damages[invalid_mask]
        warning_msg = "\n".join(
            f"Row {i}: damage_type={val}, damage_type must be in {VALID_DAMAGE_TYPE}" for i, val in zip(invalid_indices, invalid_values)
        )
        logger.warning(f"Error: Invalid damage_type values found:\n{warning_msg}")


def damagebin_tobin(stack, file_in, file_out, file_type, no_validation):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)

    if not no_validation:
        _validate(data)

    data.tofile(file_out)



#!/usr/bin/env python

from contextlib import ExitStack
import logging
import sys
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.data import TOOL_INFO

logger = logging.getLogger(__name__)


def default_toparquet(stack, file_in, file_out, file_type):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]

    file_in = resolve_file(file_in, "rb", stack)
    if file_in == sys.stdin.buffer:
        data = np.frombuffer(file_in.read(), dtype=dtype)
    else:
        data = np.fromfile(file_in, dtype=dtype)

    df = pd.DataFrame(data, columns=headers)
    table = pa.Table.from_pandas(df)
    pq.write_table(table, file_out)


def bintoparquet(file_in, file_out, file_type, **kwargs):
    """Convert bin file to parquet file based on file type
    Args:
        file_in (str | os.PathLike): Input file path
        file_out (str | os.PathLike): Output file path
        file_type (str): File type str from SUPPORTED_BINTOPARQUET
    """
    with ExitStack() as stack:
        file_out = resolve_file(file_out, "wb", stack)
        default_toparquet(stack, file_in, file_out, file_type, **kwargs)



from contextlib import ExitStack
import logging
import sys
import numba as nb
import numpy as np
from pathlib import Path

from oasislmf.pytools.common.data import (
    load_as_ndarray, nb_oasis_int,
    correlations_headers, correlations_dtype, coverages_headers,
    occurrence_dtype, occurrence_granular_dtype, periods_dtype, quantile_dtype,
    quantile_interval_dtype, returnperiods_dtype
)
from oasislmf.pytools.common.event_stream import mv_read, oasis_int, oasis_float


logger = logging.getLogger(__name__)

# Input file names (input/<file_name>)
AMPLIFICATIONS_FILE = "amplifications.bin"
CORRELATIONS_FILENAME = "correlations.bin"
COVERAGES_FILE = "coverages.bin"
EVENTRATES_FILE = "event_rates.csv"
FMPOLICYTC_FILE = "fm_policytc.bin"
FMPROGRAMME_FILE = "fm_programme.bin"
FMPROFILE_FILE = "fm_profile.bin"
FMSUMMARYXREF_FILE = "fmsummaryxref.bin"
FMXREF_FILE = "fmxref.bin"
GULSUMMARYXREF_FILE = "gulsummaryxref.bin"
ITEMS_FILE = "items.bin"
OCCURRENCE_FILE = "occurrence.bin"
PERIODS_FILE = "periods.bin"
QUANTILE_FILE = "quantile.bin"
RETURNPERIODS_FILE = "returnperiods.bin"


def read_amplifications(run_dir="", filename=AMPLIFICATIONS_FILE, use_stdin=False):
    """
    Get array of amplification IDs from amplifications.bin, where index
    corresponds to item ID.

    amplifications.bin is binary file with layout:
        reserved header (4-byte int),
        item ID 1 (4-byte int), amplification ID a_1 (4-byte int),
        ...
        item ID n (4-byte int), amplification ID a_n (4-byte int)

    Args:
        run_dir (str): path to amplifications.bin file
        filename (str | os.PathLike): amplifications file name
        use_stdin (bool): Use standard input for file data, ignores run_dir/filename. Defaults to False.
    Returns:
        items_amps (numpy.ndarray): array of amplification IDs, where index
            corresponds to item ID
    """
    header_size = 4
    if use_stdin:
        items_amps = np.frombuffer(sys.stdin.buffer.read(), dtype=np.int32, offset=header_size)
    else:
        amplification_file = Path(run_dir, filename)
        if not amplification_file.exists():
            raise FileNotFoundError('amplifications file not found.')
        items_amps = np.fromfile(amplification_file, dtype=np.int32, offset=header_size)

    # Check item IDs start from 1 and are contiguous
    if items_amps[0] != 1:
        raise ValueError(f'First item ID is {items_amps[0]}. Expected 1.')
    items_amps = items_amps.reshape(len(items_amps) // 2, 2)
    if not np.all(items_amps[1:, 0] - items_amps[:-1, 0] == 1):
        raise ValueError('Item IDs in amplifications file are not contiguous')

    items_amps = np.concatenate((np.array([0]), items_amps[:, 1]))

    return items_amps


def read_correlations(run_dir, ignore_file_type=set(), filename=CORRELATIONS_FILENAME):
    """Load the correlations from the correlations file.
    Args:
        run_dir (str): path to correlations file
        ignore_file_type (Set[str]): file extension to ignore when loading.
        filename (str | os.PathLike): correlations file name
    Returns:
        Tuple[Dict[int, int], List[int], Dict[int, int], List[Tuple[int, int]], List[int]]
        vulnerability dictionary, vulnerability IDs, areaperil to vulnerability index dictionary,
        areaperil ID to vulnerability index array, areaperil ID to vulnerability array
    """

    for ext in ["bin", "csv"]:
        if ext in ignore_file_type:
            continue

        correlations_file = Path(run_dir, filename).with_suffix("." + ext)
        if correlations_file.exists():
            logger.debug(f"loading {correlations_file}")
            if ext == "bin":
                try:
                    correlations = np.memmap(correlations_file, dtype=correlations_dtype, mode='r')
                except ValueError:
                    logger.debug("binary file is empty, numpy.memmap failed. trying to read correlations.csv.")
                    correlations = read_correlations(run_dir, ignore_file_type={'bin'}, filename=correlations_file.with_suffix(".csv").name)
            elif ext == "csv":
                # Check for header
                with open(correlations_file, "r") as fin:
                    first_line = fin.readline()
                    first_line_elements = [header.strip() for header in first_line.strip().split(',')]
                    has_header = first_line_elements == correlations_headers
                correlations = np.loadtxt(
                    correlations_file,
                    dtype=correlations_dtype,
                    delimiter=",",
                    skiprows=1 if has_header else 0,
                    ndmin=1
                )
            else:
                raise RuntimeError(f"Cannot read correlations file of type {ext}. Not Implemented.")
            return correlations

    raise FileNotFoundError(f'correlations file not found at {run_dir}. Ignoring files with ext {ignore_file_type}.')


def read_coverages(run_dir="", ignore_file_type=set(), filename=COVERAGES_FILE, use_stdin=False):
    """Load the coverages from the coverages file.
    Args:
        run_dir (str): path to coverages file
        ignore_file_type (Set[str]): file extension to ignore when loading.
        filename (str | os.PathLike): coverages file name
        use_stdin (bool): Use standard input for file data, ignores run_dir/filename. Defaults to False.
    Returns:
        numpy.array[oasis_float]: array with the coverage values for each coverage_id.
    """
    supported_exts = ["bin", "csv"]

    def read_csv_lines(lines):
        # Check for header
        first_line_elements = [header.strip() for header in lines[0].strip().split(',')]
        has_header = first_line_elements == coverages_headers
        data_lines = lines[1:] if has_header else lines
        return np.loadtxt(
            data_lines,
            dtype=oasis_float,
            delimiter=",",
            ndmin=1
        )[:, 1]

    # STDIN
    if use_stdin:
        for ext in supported_exts:
            if ext in ignore_file_type:
                continue
            if ext == "bin":
                return np.frombuffer(sys.stdin.buffer.read(), dtype=oasis_float)
            elif ext == "csv":
                lines = sys.stdin.readlines()
                return read_csv_lines(lines)
            else:
                raise RuntimeError(f"Cannot read coverages file of type {ext}. Not Implemented.")
        raise RuntimeError(
            f'coverages data not readable with use_stdin={use_stdin}, run_dir={run_dir}, filename={filename}. Ignoring files with ext {ignore_file_type}.')

    # FILE
    for ext in supported_exts:
        if ext in ignore_file_type:
            continue

        coverages_file = Path(run_dir, filename).with_suffix("." + ext)
        if not coverages_file.exists():
            continue

        if ext == "bin":
            return np.fromfile(coverages_file, dtype=oasis_float)
        elif ext == "csv":
            with ExitStack() as stack:
                fin = stack.enter_context(open(coverages_file, "r"))
                lines = fin.readlines()
            return read_csv_lines(lines)
        else:
            raise RuntimeError(f"Cannot read coverages file of type {ext}. Not Implemented.")

    raise RuntimeError(
        f'coverages data not readable with use_stdin={use_stdin}, run_dir={run_dir}, filename={filename}. Ignoring files with ext {ignore_file_type}.')


def read_event_rates(run_dir, filename=EVENTRATES_FILE):
    """Reads event rates from a CSV file
    Args:
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): event rates csv file name
    Returns:
        unique_event_ids (ndarray[oasis_int]): unique event ids
        event_rates (ndarray[oasis_float]): event rates
    """
    event_rate_file = Path(run_dir, filename)
    data = load_as_ndarray(
        run_dir,
        filename[:-4],
        np.dtype([('event_id', oasis_int), ('rate', oasis_float)]),
        must_exist=False,
        col_map={"event_id": "EventIds", "rate": "Event_rates"}
    )
    if data is None or data.size == 0:
        logger.info(f"Event rate file {event_rate_file} is empty, proceeding without event rates.")
        return np.array([], dtype=oasis_int), np.array([], dtype=oasis_float)
    unique_event_ids = data['event_id']
    event_rates = data['rate']

    # Make sure event_ids are sorted
    sort_idx = np.argsort(unique_event_ids)
    unique_event_ids = unique_event_ids[sort_idx]
    event_rates = event_rates[sort_idx]
    return unique_event_ids, event_rates


def read_quantile(sample_size, run_dir, filename=QUANTILE_FILE, return_empty=False):
    """Generate a quantile interval Dictionary based on sample size and quantile binary file
    Args:
        sample_size (int): Sample size
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): quantile binary file name
        return_empty (bool): return an empty intervals array regardless of the existence of the quantile binary
    Returns:
        intervals (quantile_interval_dtype): Numpy array emulating a dictionary for numba
    """
    intervals = []

    if return_empty:
        return np.array([], dtype=quantile_interval_dtype)
    data = load_as_ndarray(run_dir, filename[:-4], quantile_dtype, must_exist=True)
    for row in data:
        q = row["quantile"]
        # Calculate interval index and fractional part
        pos = (sample_size - 1) * q + 1
        integer_part = int(pos)
        fractional_part = pos - integer_part
        intervals.append((q, integer_part, fractional_part))

    # Convert to numpy array
    intervals = np.array(intervals, dtype=quantile_interval_dtype)
    return intervals


def read_occurrence_bin(run_dir="", filename=OCCURRENCE_FILE, use_stdin=False):
    """Read the occurrence binary file and returns an occurrence map
    Args:
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): occurrence binary file name
        use_stdin (bool): Use standard input for file data, ignores run_dir/filename. Defaults to False.
    Returns:
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
    """
    occurrence_fp = Path(run_dir, filename)
    if use_stdin:
        fin = np.frombuffer(sys.stdin.buffer.read(), dtype="u1")
    else:
        fin = np.memmap(occurrence_fp, mode="r", dtype="u1")
    cursor = 0
    valid_buff = len(fin)

    if valid_buff - cursor < np.dtype(np.int32).itemsize:
        raise RuntimeError("Error: broken occurrence file, not enough data")
    date_opts, cursor = mv_read(fin, cursor, np.int32, np.dtype(np.int32).itemsize)

    date_algorithm = date_opts & 1
    granular_date = date_opts >> 1

    # (event_id: int, period_no: int, occ_date_id: int)
    record_size = np.dtype(np.int32).itemsize * 3
    # (event_id: int, period_no: int, occ_date_id: long long)
    if granular_date:
        record_size = np.dtype(np.int32).itemsize * 2 + np.dtype(np.int64).itemsize
    # Should not get here
    if not date_algorithm and granular_date:
        raise RuntimeError("FATAL: Unknown date algorithm")

    # Extract no_of_periods
    if valid_buff - cursor < np.dtype(np.int32).itemsize:
        raise RuntimeError("Error: broken occurrence file, not enough data")
    no_of_periods, cursor = mv_read(fin, cursor, np.int32, np.dtype(np.int32).itemsize)

    num_records = (valid_buff - cursor) // record_size
    if (valid_buff - cursor) % record_size != 0:
        logger.warning(
            f"Occurrence File size (num_records: {num_records}) does not align with expected record size (record_size: {record_size})"
        )

    occ_dtype = occurrence_dtype
    if granular_date:
        occ_dtype = occurrence_granular_dtype
    occ_arr = np.zeros(0, dtype=occ_dtype)

    if num_records > 0:
        occ_arr = np.frombuffer(fin[cursor:cursor + num_records * record_size], dtype=occ_dtype)

    return occ_arr, date_algorithm, granular_date, no_of_periods


def read_occurrence(run_dir, filename=OCCURRENCE_FILE):
    """Read the occurrence binary file and returns an occurrence map
    Args:
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): occurrence binary file name
    Returns:
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
    """
    occ_arr, date_algorithm, granular_date, no_of_periods = read_occurrence_bin(run_dir, filename=filename)

    occ_dtype = occurrence_dtype
    if granular_date:
        occ_dtype = occurrence_granular_dtype

    occ_map_valtype = occ_dtype[["period_no", "occ_date_id"]]
    NB_occ_map_valtype = nb.types.Array(nb.from_dtype(occ_map_valtype), 1, "C")

    occ_map = _read_occ_arr(occ_arr, occ_map_valtype, NB_occ_map_valtype)

    return occ_map, date_algorithm, granular_date, no_of_periods


@nb.njit(cache=True, error_model="numpy")
def _read_occ_arr(occ_arr, occ_map_valtype, NB_occ_map_valtype):
    """Reads occurrence file array and returns an occurrence map of event_id to list of (period_no, occ_date_id)
    """
    occ_map = nb.typed.Dict.empty(nb_oasis_int, NB_occ_map_valtype)
    occ_map_sizes = nb.typed.Dict.empty(nb_oasis_int, nb.types.int64)
    for row in occ_arr:
        event_id = row["event_id"]
        if event_id not in occ_map:
            occ_map[event_id] = np.zeros(8, dtype=occ_map_valtype)
            occ_map_sizes[event_id] = 0
        array = occ_map[event_id]
        current_size = occ_map_sizes[event_id]

        if current_size >= len(array):  # Resize if the array is full
            new_array = np.empty(len(array) * 2, dtype=occ_map_valtype)
            new_array[:len(array)] = array
            array = new_array

        occ_map_current_size = occ_map_sizes[event_id]
        array[occ_map_current_size]["period_no"] = row["period_no"]
        array[occ_map_current_size]["occ_date_id"] = row["occ_date_id"]
        occ_map[event_id] = array
        occ_map_sizes[event_id] += 1

    for event_id in occ_map:
        occ_map[event_id] = occ_map[event_id][:occ_map_sizes[event_id]]

    return occ_map


@nb.njit(cache=True)
def occ_get_date(occ_date_id, granular_date):
    """Returns date as year, month, day, hour, minute from occ_date_id

    Args:
        occ_date_id (np.int32 | np.int64): occurrence file date id (int64 for granular dates)
        granular_date (bool): boolean for whether granular date should be extracted or not

    Returns:
        (oasis_int, oasis_int, oasis_int, oasis_int, oasis_int): Returns year, month, date, hour, minute
    """
    days = occ_date_id / (1440 - 1439 * (not granular_date))

    # Function void d(long long g, int& y, int& mm, int& dd) taken from pltcalc.cpp
    y = (10000 * days + 14780) // 3652425
    ddd = days - (365 * y + y // 4 - y // 100 + y // 400)
    if ddd < 0:
        y = y - 1
        ddd = days - (365 * y + y // 4 - y // 100 + y // 400)
    mi = (100 * ddd + 52) // 3060
    mm = (mi + 2) % 12 + 1
    y = y + (mi + 2) // 12
    dd = ddd - (mi * 306 + 5) // 10 + 1

    minutes = (occ_date_id % 1440) * granular_date
    occ_hour = minutes // 60
    occ_minutes = minutes % 60

    return y, mm, dd, occ_hour, occ_minutes


@nb.njit(cache=True)
def occ_get_date_id(granular_date, occ_year, occ_month, occ_day, occ_hour=0, occ_minute=0):
    """Returns the occ_date_id from year, month, day, hour, minute and whether it is a granular date
    Args:
        granular_date (bool): boolean for whether granular date should be extracted or not
        occ_year (int): Occurrence Year.
        occ_month (int): Occurrence Month.
        occ_day (int): Occurrence Day.
        occ_hour (int): Occurrence Hour. Defaults to 0.
        occ_minute (int): Occurrence Minute. Defaults to 0.
    Returns:
        occ_date_id (np.int64): occurrence file date id (int64 for granular dates)
    """
    occ_month = (occ_month + 9) % 12
    occ_year = occ_year - occ_month // 10
    occ_date_id = np.int64(
        365 * occ_year + occ_year // 4 - occ_year // 100 + occ_year // 400 + (occ_month * 306 + 5) // 10 + (occ_day - 1)
    )

    occ_date_id *= (1440 // (1440 - 1439 * granular_date))
    occ_date_id += (60 * occ_hour + occ_minute)
    return occ_date_id


def read_periods(no_of_periods, run_dir, filename=PERIODS_FILE):
    """Returns an array of period weights for each period between 1 and no_of_periods inclusive (with no gaps).
    Args:
        no_of_periods (int): Number of periods
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): periods binary file name
    Returns:
        period_weights (ndarray[periods_dtype]): Period weights
    """
    periods_fp = Path(run_dir, filename)

    if not periods_fp.exists():
        # If no periods binary file found, the revert to using period weights reciprocal to no_of_periods
        logger.warning(f"Periods file not found at {periods_fp}, using reciprocal calculated period weights based on no_of_periods {no_of_periods}")
        period_weights = np.array(
            [(i + 1, 1 / no_of_periods) for i in range(no_of_periods)],
            dtype=periods_dtype
        )
        return period_weights

    data = load_as_ndarray(run_dir, filename[:-4], periods_dtype, must_exist=True)
    # Less data than no_of_periods
    if len(data) != no_of_periods:
        raise RuntimeError(f"ERROR: no_of_periods does not match total period_no in period binary file {periods_fp}.")

    # Sort by period_no
    period_weights = np.sort(data, order="period_no")

    # Identify any missing periods
    expected_periods = np.arange(1, no_of_periods + 1)
    actual_periods = period_weights['period_no']
    missing_periods = np.setdiff1d(expected_periods, actual_periods)
    if len(missing_periods) > 0:
        raise RuntimeError(f"ERROR: Missing period_no in period binary file {periods_fp}.")
    return period_weights


def read_returnperiods(use_return_period_file, run_dir, filename=RETURNPERIODS_FILE):
    """Returns an array of return periods decreasing order with no duplicates.
    Args:
        use_return_period_file (bool): Bool to use Return Period File
        run_dir (str | os.PathLike): Path to input files dir
        filename (str | os.PathLike): return periods binary file name
    Returns:
        return_periods (ndarray[np.int32]): Return Periods
        use_return_period_file (bool): Bool to use Return Period File
    """
    if not use_return_period_file:
        return np.array([], dtype=returnperiods_dtype)["return_period"], use_return_period_file
    returnperiods_fp = Path(run_dir, filename)

    if not returnperiods_fp.exists():
        raise RuntimeError(f"ERROR: Return Periods file not found at {returnperiods_fp}.")

    returnperiods = load_as_ndarray(
        run_dir,
        filename[:-4],
        returnperiods_dtype,
        must_exist=False
    )

    if len(returnperiods) == 0:
        logger.warning(f"WARNING: Empty return periods file at {returnperiods_fp}. Running without defined return periods option")
        return None, False

    # Check return periods validity
    # Return periods should be unique and decreasing in order
    if len(returnperiods) != len(np.unique(returnperiods)):
        raise RuntimeError(f"ERROR: Invalid return periods file. Duplicate return periods found: {returnperiods}")
    lastrp = -1
    for rp in returnperiods["return_period"]:
        if lastrp != -1 and lastrp <= rp:
            raise RuntimeError(f"ERROR: Invalid return periods file. Non-decreasing return periods found: {returnperiods}")
        lastrp = rp

    return returnperiods["return_period"], use_return_period_file



from oasislmf.pytools.common.data import damagebin_headers, damagebin_dtype, damagebin_fmt


headers = damagebin_headers
dtype = damagebin_dtype
fmt = damagebin_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.common.data import footprint_event_headers, footprint_event_dtype, footprint_event_fmt


headers = footprint_event_headers
dtype = footprint_event_dtype
fmt = footprint_event_fmt
cli_support = ['bintocsv', 'csvtobin']



import numpy as np
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO


def amplifications_tobin(stack, file_in, file_out, file_type):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)

    # Check item IDs start from 1 and are contiguous
    if len(data) > 0 and data["item_id"][0] != 1:
        raise ValueError(f'First item ID is {data["item_id"][0]}. Expected 1.')
    if len(data) > 0 and not np.all(data["item_id"][1:] - data["item_id"][:-1] == 1):
        raise ValueError(f'Item IDs in {file_in} are not contiguous')

    # Write the 4-byte zero header
    np.array([0], dtype="i4").tofile(file_out)
    data.tofile(file_out)



"""
This file houses the classes that load the footprint data from compressed, binary, and CSV files.
"""
import json
import logging
import pickle
import mmap
import os
from contextlib import ExitStack
from typing import Dict, List, Union
from zlib import decompress

import numpy as np
import pandas as pd
import numba as nb

from oasis_data_manager.df_reader.config import clean_config, InputReaderConfig, get_df_reader
from oasis_data_manager.df_reader.reader import OasisReader
from oasis_data_manager.filestore.backends.base import BaseStorage
from .common import (
    FootprintHeader, EventIndexBin, EventIndexBinZ, Event,
    EventDynamic, footprint_filename, footprint_index_filename,
    zfootprint_filename, zfootprint_index_filename,
    csvfootprint_filename, parquetfootprint_filename,
    parquetfootprint_meta_filename, event_defintion_filename,
    hazard_case_filename, fp_format_priorities,
    parquetfootprint_chunked_dir,
    parquetfootprint_chunked_lookup, footprint_bin_lookup
)
from oasislmf.pytools.common.data import footprint_event_dtype

logger = logging.getLogger(__name__)

uncompressedMask = 1 << 1
intensityMask = 1


CURRENT_DIRECTORY = str(os.getcwd())


@nb.njit(cache=True)
def has_number_in_range(areaperil_ids, min_areaperil_id, max_areaperil_id):
    for apid in areaperil_ids:
        if min_areaperil_id <= apid <= max_areaperil_id:
            return True
    return False


def df_to_numpy(dataframe, dtype, columns={}) -> np.array:
    """

    Args:
        dataframe: DataFrame to convert to numpy
        dtype: numpy dtype of the output ndarray
        columns: optional dict-like object (with get method) mapping np_column => dataframe_column if they are different
    Returns:
        numpy nd array

    >>> dataframe = pd.DataFrame({'a':[1,2], 'b':[0.0, 1.0]})
    >>> dtype = np.dtype([('a', np.int64), ('c', np.float32),])
    >>> columns = {'c': 'b'}
    >>> df_to_numpy(dataframe, dtype, columns)
    array([(1, 0.), (2, 1.)], dtype=[('a', '<i8'), ('c', '<f4')])
    """
    numpy_data = np.empty(len(dataframe), dtype=dtype)
    for np_column in dtype.fields.keys():
        numpy_data[:][np_column] = dataframe[columns.get(np_column, np_column)].to_numpy()
    return numpy_data


@nb.njit(cache=True)
def get_event_map(event_ids):
    event_map = {}
    cur_event_id = event_ids[0]
    last_idx = 0
    for cur_idx in range(1, len(event_ids)):
        if event_ids[cur_idx] != cur_event_id:
            event_map[cur_event_id] = (last_idx, cur_idx)
            cur_event_id = event_ids[cur_idx]
            last_idx = cur_idx
    event_map[cur_event_id] = (last_idx, len(event_ids))
    return event_map


class OasisFootPrintError(Exception):
    """
    Raises exceptions when loading footprints.
    """

    def __init__(self, message: str) -> None:
        """
        The constructor of the OasisFootPrintError class.

        Args:
            message: (str) the message to be raised
        """
        super().__init__(message)


class Footprint:
    """
    This class is the base class for the footprint loaders.

    Attributes:
        storage (BaseStorage): the storage object used to lookup files
        stack (ExitStack): the context manager that combines other context managers and cleanup functions
    """

    def __init__(
            self, storage: BaseStorage,
            df_engine="oasis_data_manager.df_reader.reader.OasisPandasReader",
            areaperil_ids=None
    ) -> None:
        """
        The constructor for the Footprint class.

        Args:
            storage (BaseStorage): the storage object used to lookup files
            areaperil_ids (list): areaperil_ids that will be useful
        """
        self.storage = storage
        self.stack = ExitStack()
        self.df_engine = df_engine
        if areaperil_ids is not None:
            self.areaperil_ids = areaperil_ids
        else:
            self.areaperil_ids = None

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stack.__exit__(exc_type, exc_value, exc_traceback)

    @staticmethod
    def get_footprint_fmt_priorities():
        """
        Get list of footprint file format classes in order of priority.

        Returns: (list) footprint file format classes
        """
        format_to_class = {
            'parquet_chunk': FootprintParquetChunk,
            'parquet': FootprintParquet, 'csv': FootprintCsv,
            'binZ': FootprintBinZ, 'bin': FootprintBin,
            'parquet_dynamic': FootprintParquetDynamic,
        }
        priorities = [format_to_class[fmt] for fmt in fp_format_priorities if fmt in format_to_class]

        return priorities

    @classmethod
    def load(
        cls,
        storage: BaseStorage,
        ignore_file_type=set(),
        df_engine="oasis_data_manager.df_reader.reader.OasisPandasReader",
        areaperil_ids=None,
        **kwargs
    ):
        """
        Loads the loading classes defined in this file checking to see if the files are in the static path
        whilst doing so. The loading goes through the hierarchy with the following order:

        -> parquet
        -> compressed binary file
        -> binary file
        -> CSV file
        -> parquet (with dynamic generation)

        If the compressed binary file is present, this will be loaded. If it
        is not, then the binary file will be loaded
        and so on.

        Args:
            storage (BaseStorage): the storage object used to lookup files
            ignore_file_type (Set[str]): type of file to be skipped in the hierarchy. This can be a choice of:

            parquet
            json
            z
            bin
            idx

        Returns: (Union[FootprintBinZ, FootprintBin, FootprintCsv]) the loaded class
        """
        for footprint_class in cls.get_footprint_fmt_priorities():
            for filename in footprint_class.footprint_filenames:
                if (not storage.exists(filename) or filename.rsplit('.', 1)[-1] in ignore_file_type):
                    valid = False
                    break
            else:
                valid = True
            if valid:
                for filename in footprint_class.footprint_filenames:
                    logger.debug(f"loading {filename}")
                return footprint_class(storage, df_engine=df_engine, areaperil_ids=areaperil_ids)
        else:
            if storage.isfile("footprint.parquet"):
                raise OasisFootPrintError(
                    message="footprint.parquet needs to be partitioned in order to work, please see: "
                    "oasislmf.pytools.data_layer.conversions.footprint => convert_bin_to_parquet"
                )
            raise OasisFootPrintError(message="no valid footprint found")

    def get_event(self, event_id):
        raise NotImplementedError()

    def get_df_reader(self, filepath, **kwargs) -> OasisReader:
        # load the base df engine config and add the connection parameters
        df_reader_config = clean_config(InputReaderConfig(filepath=filepath, engine=self.df_engine))
        df_reader_config["engine"]["options"]["storage"] = self.storage

        return get_df_reader(df_reader_config, **kwargs)

    @staticmethod
    def prepare_df_data(data_frame: pd.DataFrame) -> np.array:
        """
        Reads footprint data from a parquet file.

        Returns: (np.array) footprint data loaded from the parquet file
        """
        return df_to_numpy(data_frame, Event)

    def areaperil_in_range(self, event_id, events_dict):
        if self.areaperil_ids is None:
            return True  # If its none we are searching all the places
        if event_id not in events_dict:
            return False
        min_areaperil_id, max_areaperil_id = events_dict[event_id]
        return has_number_in_range(
            self.areaperil_ids, min_areaperil_id, max_areaperil_id
        )


class FootprintCsv(Footprint):
    """
    This class is responsible for loading footprint data from CSV.

    Attributes (when in context):
        footprint (np.array[footprint_event_dtype]): event data loaded from the CSV file
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [csvfootprint_filename]

    def __enter__(self):
        self.reader = pd.read_csv()
        self.reader = self.get_df_reader("footprint.csv", dtype=footprint_event_dtype)

        self.num_intensity_bins = self.reader.query(lambda df: df['intensity_bin_id'].max())

        self.has_intensity_uncertainty = self.reader.query(
            lambda df: df.groupby(
                ['event_id', 'areaperil_id']
            ).size().max() > 1
        )

        def _fn(df):
            footprint_index_df = df.groupby('event_id', as_index=False).size()
            footprint_index_df['offset'] = (footprint_index_df['size'].cumsum() - footprint_index_df['size'])

            footprint_index_df.set_index('event_id', inplace=True)
            return footprint_index_df
        self.footprint_index = self.reader.query(_fn).to_dict('index')

        return self

    def get_event(self, event_id):
        """
        Gets the event from self.footprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[footprint_event_dtype]) the event that was extracted
        """
        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        else:
            return self.prepare_df_data(self.reader.filter(lambda df: df[df["event_id"] == event_id]).as_pandas())


class FootprintBin(Footprint):
    """
    This class is responsible loading the event data from the footprint binary files.

    Attributes (when in context):
        footprint (mmap.mmap): loaded data from the binary file which has header and then Event data
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [footprint_filename, footprint_index_filename]

    def __enter__(self):
        footprint_file = self.stack.enter_context(self.storage.with_fileno(footprint_filename))

        self.footprint = mmap.mmap(footprint_file.fileno(), length=0, access=mmap.ACCESS_READ)

        footprint_header = np.frombuffer(bytearray(self.footprint[:FootprintHeader.size]), dtype=FootprintHeader)

        self.num_intensity_bins = int(footprint_header['num_intensity_bins'])
        self.has_intensity_uncertainty = int(footprint_header['has_intensity_uncertainty'] & intensityMask)

        f = self.stack.enter_context(self.storage.with_fileno(footprint_index_filename))
        footprint_mmap = np.memmap(f, dtype=EventIndexBin, mode='r')

        self.footprint_index = pd.DataFrame(footprint_mmap, columns=footprint_mmap.dtype.names).set_index('event_id').to_dict('index')
        try:
            lookup_file = self.storage.with_fileno(footprint_bin_lookup)
            with lookup_file as f:
                lookup = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
            df = pickle.loads(lookup)

            self.events_dict = {
                row.event_id: (row.min_areaperil_id, row.max_areaperil_id)
                for row in df.itertuples(index=False)
            }
        except FileNotFoundError:
            self.events_dict = None

        return self

    def get_event(self, event_id):
        """
        Gets the event from self.footprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array(Event)) the event that was extracted
        """
        if self.events_dict:
            if not self.areaperil_in_range(event_id, self.events_dict):
                return None

        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        else:
            return np.frombuffer(self.footprint[event_info['offset']: event_info['offset'] + event_info['size']], Event)


class FootprintBinZ(Footprint):
    """
    This class is responsible for loading event data from compressed event data.

    Attributes (when in context):
        zfootprint (mmap.mmap): loaded data from the compressed binary file which has header and then Event data
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        uncompressed_size (int): the size in which the data is when it is decompressed
        index_dtype (Union[EventIndexBinZ, EventIndexBin]) the data type
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames = [zfootprint_filename, zfootprint_index_filename]

    def __enter__(self):
        zfootprint_file = self.stack.enter_context(self.storage.with_fileno(zfootprint_filename))
        self.zfootprint = mmap.mmap(zfootprint_file.fileno(), length=0, access=mmap.ACCESS_READ)

        footprint_header = np.frombuffer(bytearray(self.zfootprint[:FootprintHeader.size]), dtype=FootprintHeader)

        self.num_intensity_bins = int(footprint_header['num_intensity_bins'])
        self.has_intensity_uncertainty = int(footprint_header['has_intensity_uncertainty'] & intensityMask)
        self.uncompressed_size = int((footprint_header['has_intensity_uncertainty'] & uncompressedMask) >> 1)

        if self.uncompressed_size:
            self.index_dtype = EventIndexBinZ
        else:
            self.index_dtype = EventIndexBin

        f = self.stack.enter_context(self.storage.with_fileno(zfootprint_index_filename))

        zfootprint_mmap = np.memmap(f, dtype=self.index_dtype, mode='r')
        self.footprint_index = pd.DataFrame(zfootprint_mmap, columns=zfootprint_mmap.dtype.names).set_index('event_id').to_dict('index')

        try:
            lookup_file = self.storage.with_fileno(footprint_bin_lookup)
            with lookup_file as f:
                lookup = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
            df = pickle.loads(lookup)

            self.events_dict = {
                row.event_id: (row.min_areaperil_id, row.max_areaperil_id)
                for row in df.itertuples(index=False)
            }
        except FileNotFoundError:
            self.events_dict = None

        return self

    def get_event(self, event_id):
        """
        Gets the event from self.zfootprint based off the event ID passed in.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """

        event_info = self.footprint_index.get(event_id)
        if event_info is None:
            return
        elif self.events_dict:
            if not self.areaperil_in_range(event_id, self.events_dict):
                return None
        else:
            zdata = self.zfootprint[event_info['offset']: event_info['offset'] + event_info['size']]
            data = decompress(zdata)
            return np.frombuffer(data, Event)


class FootprintParquet(Footprint):
    """
    This class is responsible for loading event data from parquet event data.

    Attributes (when in context):
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty
        footprint_index (dict): map of footprint IDs with the index in the data
    """
    footprint_filenames: List[str] = [parquetfootprint_filename, parquetfootprint_meta_filename]

    def __enter__(self):
        with self.storage.open(parquetfootprint_meta_filename, 'r') as outfile:
            meta_data: Dict[str, Union[int, bool]] = json.load(outfile)

        self.num_intensity_bins = int(meta_data['num_intensity_bins'])
        self.has_intensity_uncertainty = int(meta_data['has_intensity_uncertainty'] & intensityMask)

        if self.areaperil_ids is not None:
            self.areaperil_ids_filter = [("areaperil_id", "in", self.areaperil_ids)]
        else:
            self.areaperil_ids_filter = None

        return self

    def get_event(self, event_id: int):
        """
        Gets the event data from the partitioned parquet data file.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """
        dir_path = f"footprint.parquet/event_id={event_id}/"
        if self.storage.exists(dir_path):
            reader = self.get_df_reader(dir_path, filters=self.areaperil_ids_filter)
            numpy_data = self.prepare_df_data(data_frame=reader.as_pandas())
            return numpy_data
        else:
            return np.empty(0, dtype=Event)


class FootprintParquetChunk(Footprint):
    footprint_filenames = [parquetfootprint_chunked_dir, parquetfootprint_meta_filename, parquetfootprint_chunked_lookup]
    current_reader = None
    current_partition = None

    def __enter__(self):
        with self.storage.open(parquetfootprint_meta_filename, 'r') as outfile:
            meta_data: Dict[str, Union[int, bool]] = json.load(outfile)

        self.num_intensity_bins = int(meta_data['num_intensity_bins'])
        self.has_intensity_uncertainty = int(meta_data['has_intensity_uncertainty'] & intensityMask)

        self.footprint_lookup_map = (self.get_df_reader("footprint_lookup.parquet").as_pandas()
                                     .set_index('event_id')
                                     .to_dict('index'))
        if self.areaperil_ids is not None:
            self.areaperil_ids_filter = [("areaperil_id", "in", list(self.areaperil_ids))]
        else:
            self.areaperil_ids_filter = None
        return self

    def get_event(self, event_id: int):
        """
        Gets the event data from the partitioned
        parquetfootprint_chunked_filename data file.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """
        event_info = self.footprint_lookup_map.get(event_id)
        if event_info is None:
            return None

        partition = event_info["partition"]
        if partition != self.current_partition:
            self.current_partition = partition
            self.current_df = self.get_df_reader(os.path.join(parquetfootprint_chunked_dir, f"footprint_{partition}.parquet"),
                                                 filters=self.areaperil_ids_filter).as_pandas()
            if len(self.current_df):
                self.event_map = get_event_map(self.current_df["event_id"].to_numpy())
            else:
                self.event_map = {}

        if event_id in self.event_map:
            start_idx, end_idx = self.event_map[event_id]
            return self.prepare_df_data(data_frame=self.current_df[start_idx: end_idx])
        else:
            return None

    def get_events(self, partition):
        reader = self.get_df_reader(os.path.join(parquetfootprint_chunked_dir, f"footprint_{partition}.parquet"))

        df = reader.as_pandas()
        events = [self.prepare_df_data(data_frame=group) for _, group in df.groupby("event_id")]
        return events


class FootprintParquetDynamic(Footprint):
    """
    This class is responsible for loading event data from parquet dynamic event sets and maps
    It will build the footprint from the underlying event defintion and hazard case files

    Attributes (when in context):
        num_intensity_bins (int): number of intensity bins in the data
        has_intensity_uncertainty (bool): if the data has uncertainty. Only "no" is supported
        return periods (list): the list of return periods in the model. Not currently used
    """
    footprint_filenames: List[str] = [event_defintion_filename, hazard_case_filename, parquetfootprint_meta_filename]

    def __enter__(self):
        with self.storage.open(parquetfootprint_meta_filename, 'r') as outfile:
            meta_data: Dict[str, Union[int, bool]] = json.load(outfile)

        self.num_intensity_bins = int(meta_data['num_intensity_bins'])
        self.has_intensity_uncertainty = int(meta_data['has_intensity_uncertainty'] & intensityMask)

        self.df_location_sections = pd.read_csv('input/sections.csv')
        self.location_sections = set(list(self.df_location_sections['section_id']))
        if self.areaperil_ids is None:
            self.areaperil_ids = pd.read_csv('input/keys.csv', usecols=['AreaPerilID']).AreaPerilID.unique()

        return self

    def get_event(self, event_id: int):
        """
        Gets the event data from the partitioned parquet data file.

        Args:
            event_id: (int) the ID belonging to the Event being extracted

        Returns: (np.array[Event]) the event that was extracted
        """
        event_defintion_reader = self.get_df_reader(event_defintion_filename, filters=[("event_id", "==", event_id)])
        df_event_defintion = event_defintion_reader.as_pandas()
        event_sections = list(df_event_defintion['section_id'])
        sections = list(set(event_sections) & self.location_sections)

        if len(sections) > 0:
            df_hazard_case = {}
            for section in sections:
                hazard_case_reader = self.get_df_reader(
                    f'{hazard_case_filename}/section_id={int(section)}',
                    filters=[("areaperil_id", "in", self.areaperil_ids)]
                )
                df_hazard_case[section] = hazard_case_reader.as_pandas()
                df_hazard_case[section]['section_id'] = section
            df_hazard_case = pd.concat(df_hazard_case, ignore_index=True)

            from_cols = ['section_id', 'areaperil_id', 'intensity']
            to_cols = from_cols + ['interpolation', 'return_period']

            df_hazard_case_from = df_hazard_case.merge(
                df_event_defintion, left_on=['section_id', 'return_period'], right_on=['section_id', 'rp_from'])[from_cols].rename(
                    columns={'intensity': 'from_intensity'})

            df_hazard_case_to = df_hazard_case.merge(
                df_event_defintion, left_on=['section_id', 'return_period'], right_on=['section_id', 'rp_to'])[to_cols].rename(
                    columns={'intensity': 'to_intensity'})

            df_footprint = df_hazard_case_from.merge(df_hazard_case_to, on=['section_id', 'areaperil_id'], how='outer')
            df_footprint['from_intensity'] = df_footprint['from_intensity'].fillna(0)

            if len(df_footprint.index) > 0:
                df_footprint['intensity'] = np.floor(df_footprint.from_intensity + (
                    (df_footprint.to_intensity - df_footprint.from_intensity) * df_footprint.interpolation))
                df_footprint['intensity'] = df_footprint['intensity'].astype('int')
                df_footprint = df_footprint.sort_values('intensity', ascending=False)
                df_footprint = df_footprint.drop_duplicates(subset=['areaperil_id'], keep='first')
                df_footprint['intensity_bin_id'] = 0  # Placeholder for intensity bin ID
                df_footprint['probability'] = 1
            else:
                df_footprint.loc[:, 'intensity'] = []
                df_footprint.loc[:, 'intensity_bin_id'] = []
                df_footprint.loc[:, 'probability'] = []

            numpy_data = np.empty(len(df_footprint), dtype=EventDynamic)
            for column in ['areaperil_id', 'intensity_bin_id', 'intensity', 'probability', 'return_period']:
                numpy_data[:][column] = df_footprint[column].to_numpy()

            return numpy_data


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from oasislmf.pytools.common.data import fm_profile_headers, fm_profile_dtype, fm_profile_fmt


headers = fm_profile_headers
dtype = fm_profile_dtype
fmt = fm_profile_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.elt.data import SELT_headers, SELT_dtype, SELT_fmt


headers = SELT_headers
dtype = SELT_dtype
fmt = SELT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']
